(*
 * Coq code for "A Translation from Typed Assebmly Language to
 *                                  Certified Assembly Programming"
 *
 * Axioms used in the implementatio
 *
 * (for Coq version 8)
 *)

Require Export Eqdep.
Require Import Map.
Require Import Print tm .

(* Extensional equality on functions, supposed to be a "safe" extention of Coq,
   i.e., does not violate the consistency of Coq. *)
Axiom ext_eq : forall A B (f g : A -> B), (forall x, f x = g x) -> f = g.

(* Assumption about heap being finite at any given time.
   There are two ways to remove the axiom, either by changing the encoding
   of mapping into a finite one, or use the following assertion function to
   wrap the actural assertion, which should be easy to satisfy.

Definition FiniteHeapAssertion (P:Assertion) := fun S =>
  Ex l. P (fH (_H S) l, _R S) ./\
        <<Map.in_dom (_H S) l /\ forall l', l' > l -> Map.not_in_dom (_H S) l'>>.

*)
Axiom FiniteHeap : forall H:Heap, exists l,
                     Map.in_dom H l /\ forall l', l' > l -> Map.not_in_dom H l'.


