(*
 * Coq code for "A Translation from Typed Assebmly Language to
 *                                  Certified Assembly Programming"
 *
 * type list
 *
 * (for Coq version 8)
 *)

Inductive TyList : Type :=
  | tO : TyList
  | tS : Type -> TyList -> TyList.

Fixpoint tplus (L L' : TyList) {struct L} : TyList :=
  match L with
  | tO => L'
  | tS A L => tS A (tplus L L')
  end.

Lemma eq_tS_head : forall A A' L L', tS A L = tS A' L' -> A = A'.
intros.
inversion H.
auto.
Defined.

Lemma eq_tS_tail : forall A A' L L', tS A L = tS A' L' -> L = L'.
intros.
inversion H.
auto.
Defined.

Lemma eq_tS : forall A L L', L = L' -> tS A L = tS A L'.
intros.
inversion H.
auto.
Defined.

Lemma eq_tS_head_eq : forall A A' L, A = A' -> tS A L = tS A' L.
intros.
rewrite H.
auto.
Defined.

Lemma eq_tplus : forall L L' L'', L = L' -> tplus L L'' = tplus L' L''.
intros.
rewrite H.
auto.
Defined.

Lemma eq_tplus_tail : forall L L' L'', L = L' -> tplus L'' L = tplus L'' L'.
intros.
rewrite H.
auto.
Defined.

Lemma eq_tplus_tO : forall L, L = tplus tO L.
simpl in |- *.
auto.
Defined.

Lemma eq_tplus_tO_tail : forall L, L = tplus L tO.
simple induction L.
 simpl in |- *.
   auto.
 simpl in |- *.
   intros.
   rewrite H.
   simpl in |- *.
   rewrite <- H.
   rewrite <- H.
   auto.
Defined.

Lemma eq_tS_tplus : forall A L L', tS A (tplus L L') = tplus (tS A L) L'.
simpl in |- *.
auto.
Defined.

Lemma eq_tplus_tS :
 forall A L L', tplus L (tS A L') = tplus (tplus L (tS A tO)) L'.
intros.
induction  L as [| T L HrecL].
 simpl in |- *.
   auto.
 simpl in |- *.
   rewrite HrecL.
   auto.
Defined.

Lemma eq_tplus_tplus :
 forall L L' L'', tplus (tplus L L') L'' = tplus L (tplus L' L'').
intros.
induction  L as [| T L HrecL].
 simpl in |- *.
   auto.
 simpl in |- *.
   rewrite HrecL.
   auto.
Defined.

Lemma eq_tS_tplus_eq : forall A L, tS A L = tplus (tS A tO) L.
simpl in |- *.
auto.
Defined.