(*
 * Coq code for "A Translation from Typed Assebmly Language to
 *                                  Certified Assembly Programming"
 *
 * XCAP with logical recursive specifications and logical mutable reference
 *
 * (for Coq version 8)
 *)

Require Import Eqdep.
Require Import natutil.
Require Import Map.
Require Import Mapt.
Require Import tm.
Require Import tylist.
Require Import axiom.

Inductive PropX : TyList -> Type :=
  | var  : forall L A, A                   -> PropX (tS A L)
  | lift : forall L A, PropX L             -> PropX (tS A L)
  | cptr : forall L,   Word -> (State -> PropX L) -> PropX L
  | ref  : forall L,   Word -> (Word  -> PropX L) -> PropX L
  | prop : forall L,   Prop                       -> PropX L
  | andx : forall L,   PropX L -> PropX L         -> PropX L
  | orx  : forall L,   PropX L -> PropX L         -> PropX L
  | impx : forall L,   PropX L -> PropX L         -> PropX L
  | allx : forall L A, (A -> PropX L)             -> PropX L
  | extx : forall L A, (A -> PropX L)             -> PropX L
  | allv : forall L A, PropX (tplus L (tS A tO))  -> PropX L
  | extv : forall L A, PropX (tplus L (tS A tO))  -> PropX L
  | mu   : forall L A, (A -> (PropX (tplus L (tS A tO)))) -> A -> PropX L.

Notation "<< p >>" := (prop _ p  ) (at level 40).
Notation "P ./\ Q" := (andx _ P Q) (at level 51, right associativity).
Notation "P .\/ Q" := (orx  _ P Q) (at level 70, right associativity).
Notation "P .-> Q" := (impx _ P Q) (at level 100, right associativity).

Notation "'All' x1 . P" := (allx _ _ (fun x1 => P)) (at level 110).
Notation "'Ex' x1 . P" := (extx _ _ (fun x1 => P)) (at level 110).
Notation "'Ex' x1 , x2 . P" := (Ex x1 . (Ex x2 . P)) (at level 110).
Notation "'Ex' x1 , x2 , x3 . P" := (Ex x1, x2 . (Ex x3 . P)) (at level 110).
Notation "'Ex' x1 , x2 , x3 , x4 . P" :=
                                    (Ex x1, x2, x3 . (Ex x4 . P)) (at level 110).
Notation "'Ex' x1 , x2 , x3 , x4 , x5 . P" :=
                                (Ex x1, x2, x3, x4 . (Ex x5 . P)) (at level 110).
Notation "'Ex' x1 , x2 , x3 , x4 , x5 , x6 . P" :=
                            (Ex x1, x2, x3, x4, x5 . (Ex x6 . P)) (at level 110).
Notation "'Ex' x1 , x2 , x3 , x4 , x5 , x6 , x7 . P" :=
                        (Ex x1, x2, x3, x4, x5, x6 . (Ex x7 . P)) (at level 110).
Notation "'Ex' x1 , x2 , x3 , x4 , x5 , x6 , x7 , x8 . P" :=
                    (Ex x1, x2, x3, x4, x5, x6, x7 . (Ex x8 . P)) (at level 110).
Notation "'Ex' x1 , x2 , x3 , x4 , x5 , x6 , x7 , x8 , x9 . P" :=
                (Ex x1, x2, x3, x4, x5, x6, x7, x8 . (Ex x9 . P)) (at level 110).

Definition AssertionX L := State -> PropX L.

Definition Assertion := AssertionX tO.

Definition WordTy := Word -> PropX tO.

Definition CdHpSpec := Map Label Assertion.

Definition DtHpSpec := Map Label WordTy.

Fixpoint Lift (tl tl' : TyList) (P : PropX tl') {struct tl} :
 PropX (tplus tl tl') :=
  match tl return (PropX (tplus tl tl')) with
  | tO =>                                                                       eq_rect _ _
          P                                                                     _ (eq_tplus_tO _)
  | tS A tl =>                                                                  eq_rect _ _
               (lift _ _ (Lift tl tl' P))                                       _ (eq_tS_tplus _ _ _)
  end.

Fixpoint Shift L L' (p : PropX (tplus L L')) A {struct p} :
                                                    PropX (tplus L (tS A L')) :=
  match p                                                                       in (PropX L'') return (L'' = tplus L L' -> PropX (tplus L (tS A L')))
  with
  | var _ _ v =>                                                                fun pf =>
      match L                                                                   as L'' return (L = L'' -> PropX (tplus L (tS A L')))
      with
      | tO =>                                                                   fun pf' => eq_rect _ _
              (lift _ _                                                         (eq_rect _ _
                    (var _ _ v)                                                 _ (trans_eq pf (eq_tplus _ _ _ pf')))
              )                                                                 _ (trans_eq (eq_tplus_tO _) (eq_tplus _ _ _ (sym_eq pf')))
      | _  =>                                                                   fun pf' => eq_rect _ _
              (var _ _ v)                                                       _ (trans_eq (eq_tS_tplus _ _ _)
                                                                                (eq_tplus _ _ _
                                                                                   (trans_eq
                                                                                      (eq_tS_head_eq _ _ _
                                                                                         (eq_tS_head _ _ _ _
                                                                                            (trans_eq (trans_eq pf (eq_tplus _ _ _ pf'))
                                                                                               (sym_eq (eq_tS_tplus _ _ _)))))
                                                                                      (sym_eq pf'))))
      end                                                                       (refl_equal _)
  | lift _ _ p =>                                                               fun pf =>
      match L                                                                   as L'' return (L = L'' -> PropX (tplus L (tS A L')))
      with
      | tO =>                                                                   fun pf' => eq_rect _ _
              (lift _ _                                                         (eq_rect _ _
                    (lift _ _ p)                                                _ (trans_eq pf (eq_tplus _ _ _ pf')))
              )                                                                 _ (trans_eq (eq_tplus_tO _) (eq_tplus _ _ _ (sym_eq pf')))
      | _  =>                                                                   fun pf' => eq_rect _ _
              (lift _ _ (Shift _ _                                              (eq_rect _ _
                               p                                                _ (eq_tS_tail _ _ _ _
                                                                                (trans_eq (trans_eq pf (eq_tplus _ _ _ pf'))
                                                                                (sym_eq (eq_tS_tplus _ _ _)))))
                               A)
                        )                                                       _ (trans_eq (eq_tS_tplus _ _ _) (eq_tplus _ _ _ (sym_eq pf')))
      end                                                                       (refl_equal _)
  | cptr _ l P =>                                                               fun pf =>
                  cptr _ l (fun x => Shift _ _                                  (eq_rect _ _
                                           (P x)                                _ pf)
                                           A)
  | ref _ l P =>                                                               fun pf =>
                  ref _ l (fun x => Shift _ _                                  (eq_rect _ _
                                           (P x)                                _ pf)
                                           A)
  | prop _ p  =>                                                                fun pf =>
                 << p >>
  | andx _ p q =>                                                               fun pf =>
                  Shift _ _                                                     (eq_rect _ _
                        p                                                       _ pf)
                        A
              ./\ Shift _ _                                                     (eq_rect _ _
                        q                                                       _ pf)
                        A
  | orx _ p q =>                                                                fun pf =>
                  Shift _ _                                                     (eq_rect _ _
                        p                                                       _ pf)
                        A
              .\/ Shift _ _                                                     (eq_rect _ _
                        q                                                       _ pf)
                        A
  | impx _ p q =>                                                               fun pf =>
                  Shift _ _                                                     (eq_rect _ _
                        p                                                       _ pf)
                        A
              .-> Shift _ _                                                     (eq_rect _ _
                        q                                                       _ pf)
                        A
  | allx _ _ P =>                                                               fun pf =>
                  allx _ _ (fun x => Shift _ _                                  (eq_rect _ _
                                           (P x)                                _ pf)
                                           A)
  | extx _ _ P =>                                                               fun pf =>
                  Ex x . Shift _ _                                              (eq_rect _ _
                               (P x)                                            _ pf)
                               A
  | allv _ _ p =>                                                               fun pf =>
                  allv _ _                                                      (eq_rect _ _
                       (Shift _ _                                               (eq_rect _ _
                              p                                                 _ (trans_eq (eq_tplus _ _ _ pf) (eq_tplus_tplus _ _ _)))
                              _)                                                _ (trans_eq (eq_tplus_tail _ _ _ (eq_tS_tplus _ _ _))
                                                                                    (sym_eq (eq_tplus_tplus _ _ _))))
  | extv _ _ p =>                                                               fun pf =>
                  extv _ _                                                      (eq_rect _ _
                       (Shift _ _                                               (eq_rect _ _
                              p                                                 _ (trans_eq (eq_tplus _ _ _ pf) (eq_tplus_tplus _ _ _)))
                              _)                                                _ (trans_eq (eq_tplus_tail _ _ _ (eq_tS_tplus _ _ _))
                                                                                (sym_eq (eq_tplus_tplus _ _ _))))
  | mu _ _ P v => fun pf => mu _ _ (fun x => eq_rect _ _ (Shift _ _
                                        (eq_rect _ _ (P x)
                                                       _ (trans_eq
                                                           (eq_tplus _ _ _ pf)
                                                           (eq_tplus_tplus _ _ _)))
                                                       _)
                                  _ (trans_eq
                                      (eq_tplus_tail _ _ _ (eq_tS_tplus _ _ _))
                                      (sym_eq (eq_tplus_tplus _ _ _))))
                              v
  end                                                                           (refl_equal _)
.

Fixpoint Subst L A (p : PropX (tS A L)) {struct p}:
 (A -> PropX L) -> PropX L :=
  match p                                                                       in (PropX L') return (L' = tS A L -> (A -> PropX L) -> PropX L)
  with
  | var _ _ v  => fun pf Q => Q (eq_rect _ (fun T => T) v _ (eq_tS_head _ _ _ _ pf))
  | lift _ _ p => fun pf _ => eq_rect _ _ p _ (eq_tS_tail _ _ _ _ pf)
  | cptr _ l P =>
      fun pf Q => cptr _ l (fun x => Subst _ _ (eq_rect _ _ (P x) _ pf) Q)
  | ref  _ l P =>
      fun pf Q => ref  _ l (fun x => Subst _ _ (eq_rect _ _ (P x) _ pf) Q)
  | prop _ p => fun _ Q => << p >>
  | andx _ p q =>
      fun pf Q =>
      Subst _ _ (eq_rect _ _ p _ pf) Q ./\ Subst _ _ (eq_rect _ _ q _ pf) Q
  | orx _ p q =>
      fun pf Q =>
      Subst _ _ (eq_rect _ _ p _ pf) Q .\/ Subst _ _ (eq_rect _ _ q _ pf) Q
  | impx _ p q =>
      fun pf Q =>
      Subst _ _ (eq_rect _ _ p _ pf) Q .-> Subst _ _ (eq_rect _ _ q _ pf) Q
  | allx _ _ P =>
      fun pf Q => allx _ _ (fun x => Subst _ _ (eq_rect _ _ (P x) _ pf) Q)
  | extx _ _ P => fun pf Q => Ex x . Subst _ _ (eq_rect _ _ (P x) _ pf) Q
  | allv _ _ p =>
      fun pf Q =>
      allv _ _
        (Subst _ _
           (eq_rect _ _ p _
              (trans_eq (eq_tplus _ _ _ pf) (sym_eq (eq_tS_tplus _ _ _))))
           (fun x => Shift _ _ (eq_rect _ _ (Q x) _ (eq_tplus_tO_tail _)) _))
  | extv _ _ p =>
      fun pf Q =>
      extv _ _
        (Subst _ _
           (eq_rect _ _ p _
              (trans_eq (eq_tplus _ _ _ pf) (sym_eq (eq_tS_tplus _ _ _))))
           (fun x => Shift _ _ (eq_rect _ _ (Q x) _ (eq_tplus_tO_tail _)) _))
  | mu _ _ P v => fun pf Q => mu _ _ (fun x => Subst _ _ (eq_rect _ _ (P x)
                                               _ (trans_eq
                                                   (eq_tplus _ _ _ pf)
                                                   (sym_eq (eq_tS_tplus _ _ _))))
                                              (fun x => Shift _ _ (eq_rect _ _ (Q x)
                                                           _ (eq_tplus_tO_tail _))
                                                          _))
                                  v
  end (refl_equal _).

Inductive Env : Type :=
  | nil : Env
  | cons : (PropX tO) -> Env -> Env.

Fixpoint InE (E : Env) (p : PropX tO) {struct E} : Prop :=
  match E with
  | nil => False
  | cons q E => p = q \/ InE E p
  end.

Lemma eq_tS_T : forall T T' t t', tS T t = tS T' t' -> T = T'.
intros. inversion H. auto.
Defined.

Lemma eq_tS_t : forall T T' t t', tS T t = tS T' t' -> t = t'.
intros. inversion H. auto.
Defined.

Lemma eq_tplus_full : forall T T' t t', T=T' -> t=t' -> tplus T t = tplus T' t'.
intros. rewrite H. rewrite H0. auto.
Defined.

Inductive OK : Env -> CdHpSpec -> DtHpSpec -> PropX tO -> Prop :=
  | ok_env    : forall E Si Fi p, InE E p -> OK E Si Fi p
  | ok_cptr_i : forall E Si Fi l P, lookup Si l P -> OK E Si Fi (cptr _ l P)
  | ok_cptr_e : forall E Si Fi l P q,
                  OK E Si Fi (cptr _ l P) ->
                  (lookup Si l P -> OK E Si Fi q) ->
                  OK E Si Fi q
  | ok_ref_i : forall E Si Fi l P, lookup Fi l P -> OK E Si Fi (ref _ l P)
  | ok_ref_e : forall E Si Fi l P q,
                  OK E Si Fi (ref _ l P) ->
                  (lookup Fi l P -> OK E Si Fi q) ->
                  OK E Si Fi q
  | ok_prop_i : forall E Si Fi (p : Prop), p -> OK E Si Fi (<< p >>)
  | ok_prop_e : forall E Si Fi (p : Prop) q,
                  OK E Si Fi (<< p >>) -> (p -> OK E Si Fi q) -> OK E Si Fi q
  | ok_and_i  : forall E Si Fi p q, OK E Si Fi p -> OK E Si Fi q -> OK E Si Fi (p ./\ q)
  | ok_and_e1 : forall E Si Fi p q, OK E Si Fi (p ./\ q) -> OK E Si Fi p
  | ok_and_e2 : forall E Si Fi p q, OK E Si Fi (p ./\ q) -> OK E Si Fi q
  | ok_or_i1  : forall E Si Fi p q, OK E Si Fi p -> OK E Si Fi (p .\/ q)
  | ok_or_i2  : forall E Si Fi p q, OK E Si Fi q -> OK E Si Fi (p .\/ q)
  | ok_or_e   : forall E Si Fi p q r,
                  OK E Si Fi (p .\/ q) ->
                  OK (cons p E) Si Fi r -> OK (cons q E) Si Fi r -> OK E Si Fi r
  | ok_imp_i  : forall E Si Fi p q, OK (cons p E) Si Fi q -> OK E Si Fi (p .-> q)
  | ok_imp_e  : forall E Si Fi p q, OK E Si Fi p -> OK E Si Fi (p .-> q) -> OK E Si Fi q
  | ok_allx_i : forall E Si Fi A (P : A -> _),
                  (forall x, OK E Si Fi (P x)) -> OK E Si Fi (allx _ _ P)
  | ok_allx_e : forall E Si Fi A P, OK E Si Fi (allx _ _ P)
                                                  -> forall B : A, OK E Si Fi (P B)
  | ok_extx_i : forall E Si Fi A (P : A -> _) x,
                  OK E Si Fi (P x) -> OK E Si Fi (extx _ _ P)
  | ok_extx_e : forall E Si Fi A P q,
                  OK E Si Fi (extx _ _ P) ->
                    (forall B : A, OK (cons (P B) E) Si Fi q) -> OK E Si Fi q
  | ok_allv_i : forall E Si Fi A p,
                  (forall x : A -> _, OK E Si Fi (Subst _ _ p x))
                                                        -> OK E Si Fi (allv tO _ p)

  | ok_extv_i : forall E Si Fi A p (q:A->_),
                  OK E Si Fi (Subst _ _ p q)               -> OK E Si Fi (extv tO _ p)

  | ok_mu_i   : forall E Si Fi A p (v:A),
                  OK E Si Fi (Subst _ _ (p v) (fun v => mu tO _ p v))
                    -> OK E Si Fi (mu tO _ p v)
.

Inductive OB : Env -> CdHpSpec -> DtHpSpec -> PropX tO -> bool -> Prop :=
  | ob_coer   : forall E Si Fi p, OB E Si Fi p false -> OB E Si Fi p true
  | ob_env    : forall E Si Fi p, InE E p -> OB E Si Fi p false
  | ob_cptr_i : forall E Si Fi l P, lookup Si l P -> OB E Si Fi (cptr _ l P) true
  | ob_cptr_e : forall E Si Fi l P q, OB E Si Fi (cptr _ l P) false ->
                                   (lookup Si l P -> OB E Si Fi q true) ->
                                   OB E Si Fi q true
  | ob_ref_i  : forall E Si Fi l P, lookup Fi l P -> OB E Si Fi (ref _ l P) true
  | ob_ref_e  : forall E Si Fi l P q, OB E Si Fi (ref _ l P) false ->
                                   (lookup Fi l P -> OB E Si Fi q true) ->
                                   OB E Si Fi q true
  | ob_prop_i : forall E Si Fi (p : Prop), p -> OB E Si Fi (<< p >>) true
  | ob_prop_e : forall E Si Fi (p : Prop) q, OB E Si Fi (<< p >>) false ->
                                          (p -> OB E Si Fi q true) ->
                                          OB E Si Fi q true
  | ob_and_i  : forall E Si Fi p q, OB E Si Fi p true -> OB E Si Fi q true ->
                                                   OB E Si Fi (p ./\ q) true
  | ob_and_e1 : forall E Si Fi p q, OB E Si Fi (p ./\ q) false -> OB E Si Fi p false
  | ob_and_e2 : forall E Si Fi p q, OB E Si Fi (p ./\ q) false -> OB E Si Fi q false
  | ob_or_i1  : forall E Si Fi p q, OB E Si Fi p true -> OB E Si Fi (p .\/ q) true
  | ob_or_i2  : forall E Si Fi p q, OB E Si Fi q true -> OB E Si Fi (p .\/ q) true
  | ob_or_e   : forall E Si Fi p q r, OB E Si Fi (p .\/ q) false ->
                                   OB (cons p E) Si Fi r true ->
                                   OB (cons q E) Si Fi r true ->
                                   OB E Si Fi r true
  | ob_imp_i  : forall E Si Fi p q, OB (cons p E) Si Fi q true ->
                                 OB E Si Fi (p .-> q) true
  | ob_imp_e  : forall E Si Fi p q, OB E Si Fi p true -> OB E Si Fi (p .-> q) false ->
                                                   OB E Si Fi q false
  | ob_allx_i : forall E Si Fi A (P : A -> _),
                   (forall x, OB E Si Fi (P x) true) -> OB E Si Fi (allx _ _ P) true
  | ob_allx_e : forall E Si Fi A P, OB E Si Fi (allx _ _ P) false
                                            -> forall B : A, OB E Si Fi (P B) false
  | ob_extx_i : forall E Si Fi A (P : A -> _) x,
                   OB E Si Fi (P x) true -> OB E Si Fi (extx _ _ P) true
  | ob_extx_e : forall E Si Fi A P q,
                   OB E Si Fi (extx _ _ P) false ->
                   (forall B : A, OB (cons (P B) E) Si Fi q true) -> OB E Si Fi q true
  | ob_allv_i : forall E Si Fi A p,
                  (forall x : A -> _, OB E Si Fi (Subst _ _ p x) true)
                                                   -> OB E Si Fi (allv tO _ p) true
  | ob_extv_i : forall E Si Fi A p (q:A->_),
                  OB E Si Fi (Subst _ _ p q) true     -> OB E Si Fi (extv tO _ p) true
  | ob_mu_i   : forall E Si Fi A p (v:A),
                  OB E Si Fi (Subst _ _ (p v) (fun v => mu tO _ p v)) true
                    -> OB E Si Fi (mu tO _ p v) true
                  .

Theorem Soundness_of_Normal_Deductions :
  forall E Si Fi p b, OB E Si Fi p b -> OK E Si Fi p.
induction 1.
auto.
apply ok_env. auto.
apply ok_cptr_i. auto.
apply ok_cptr_e with l P. auto. auto.
apply ok_ref_i. auto.
apply ok_ref_e with l P. auto. auto.
apply ok_prop_i. auto.
apply ok_prop_e with p. auto. auto.
apply ok_and_i. auto. auto.
apply ok_and_e1 with q. auto. auto.
apply ok_and_e2 with p. auto. auto.
apply ok_or_i1. auto.
apply ok_or_i2. auto.
apply ok_or_e with p q. auto. auto. auto.
apply ok_imp_i. auto.
apply ok_imp_e with p. auto. auto.
apply ok_allx_i. auto.
apply ok_allx_e with A. auto.
apply ok_extx_i with x. auto.
apply ok_extx_e with A P. auto. auto.
apply ok_allv_i. auto.
apply ok_extv_i with q. auto.
apply ok_mu_i. auto.
Qed.

Lemma Empty_Env_Deduce_No_Extrations :
  forall E Si Fi p b, OB E Si Fi p b -> E=nil -> b=false -> False.
induction 1; intros; try tauto; try inversion H1.
rewrite H0 in H. inversion H.
inversion H2.
inversion H2.
Qed.

Theorem Validity_of_Normal_Deductions :
  forall E Si Fi p b, OB E Si Fi p b -> E=nil -> match p in PropX t return t=tO -> Prop with
                                   | var _ _ _  => fun pf => False
                                   | lift _ _ _ => fun pf => False
                                   | cptr _ l P => fun pf => lookup Si l (fun S => eq_rect _ _ (P S) _ pf)
                                   | ref  _ l P => fun pf => lookup Fi l (fun S => eq_rect _ _ (P S) _ pf)
                                   | prop _ p   => fun pf => p
                                   | andx _ p q => fun pf => OB nil Si Fi (eq_rect _ _ p _ pf) true /\ OB nil Si Fi (eq_rect _ _ q _ pf) true
                                   | orx _ p q  => fun pf => OB nil Si Fi (eq_rect _ _ p _ pf) true \/ OB nil Si Fi (eq_rect _ _ q _ pf) true
                                   | impx _ p q => fun pf => OB (cons (eq_rect _ _ p _ pf) nil) Si Fi (eq_rect _ _ q _ pf) true
                                   | allx _ A P => fun pf => forall x, OB nil Si Fi (eq_rect _ _ (P x) _ pf) true
                                   | extx _ A P => fun pf => exists x, OB nil Si Fi (eq_rect _ _ (P x) _ pf) true
                                   | allv _ A p => fun pf => forall x, OB nil Si Fi (Subst _ _
                                     (eq_rect _ _ p _
                                            (trans_eq (eq_tplus _ _ _ pf)
                                                      (sym_eq (eq_tplus_tO _))))
                                     x) true
                                   | extv _ A p => fun pf => exists x, OB nil Si Fi (Subst _ _
                                     (eq_rect _ _ p _
                                            (trans_eq (eq_tplus _ _ _ pf)
                                                      (sym_eq (eq_tplus_tO _))))
                                     x) true
                                   | mu _ _ P v => fun pf => OB nil Si Fi (Subst _ _
                                   (eq_rect _ _ (P v)
                                               _ (trans_eq (eq_tplus _ _ _ pf)
                                                      (sym_eq (eq_tplus_tO _))))
                                                      (fun v => mu tO _ (fun v => (eq_rect _ _ (P v)
                                               _ (eq_tplus _ _ _ pf))) v)) true
                                   end (refl_equal _).
induction 1; intros; try rewrite <- H1; try tauto; try auto.
rewrite H0 in H. inversion H.
simpl. rewrite <- (ext_eq _ _ P (fun S => P S)). auto. auto.
simpl in IHOB. rewrite <- (ext_eq _ _ P (fun S => P S)) in IHOB. auto. auto. tauto. auto.
simpl. rewrite <- (ext_eq _ _ P (fun S => P S)). auto. auto.
simpl in IHOB. rewrite <- (ext_eq _ _ P (fun S => P S)) in IHOB. auto. auto. tauto. auto.
simpl in IHOB. generalize (Empty_Env_Deduce_No_Extrations _ _ _ _ _ H H0 (refl_equal _)). tauto.
simpl in IHOB. generalize (Empty_Env_Deduce_No_Extrations _ _ _ _ _ H H0 (refl_equal _)). tauto.
simpl. rewrite <- H0. left. auto.
simpl. rewrite <- H0. right. auto.
simpl in IHOB1. generalize (Empty_Env_Deduce_No_Extrations _ _ _ _ _ H H2 (refl_equal _)). tauto.
simpl. rewrite <- H0. auto.
simpl in IHOB2. generalize (Empty_Env_Deduce_No_Extrations _ _ _ _ _ H0 H1 (refl_equal _)). tauto.
simpl. apply H.
simpl in IHOB. generalize (Empty_Env_Deduce_No_Extrations _ _ _ _ _ H H0 (refl_equal _)). tauto.
simpl. exists x. rewrite <- H0. auto.
simpl in IHOB. generalize (Empty_Env_Deduce_No_Extrations _ _ _ _ _ H H2 (refl_equal _)). tauto.
simpl. exists q. rewrite <- H0. auto.
simpl. rewrite <- (ext_eq _ _ p (fun v => p v)). rewrite <- H0. auto. auto.
Qed.

Inductive OP : Env -> CdHpSpec -> DtHpSpec -> PropX tO -> bool -> Prop :=
  | op_coer   : forall E Si Fi p, OP E Si Fi p false -> OP E Si Fi p true
  | op_coer'  : forall E Si Fi p, OP E Si Fi p true -> OP E Si Fi p false
  | op_env    : forall E Si Fi p, InE E p -> OP E Si Fi p false
  | op_cptr_i : forall E Si Fi l P, lookup Si l P -> OP E Si Fi (cptr _ l P) true
  | op_cptr_e : forall E Si Fi l P q, OP E Si Fi (cptr _ l P) false ->
                                   (lookup Si l P -> OP E Si Fi q true) ->
                                   OP E Si Fi q true
  | op_ref_i : forall E Si Fi l P, lookup Fi l P -> OP E Si Fi (ref _ l P) true
  | op_ref_e : forall E Si Fi l P q, OP E Si Fi (ref _ l P) false ->
                                   (lookup Fi l P -> OP E Si Fi q true) ->
                                   OP E Si Fi q true
  | op_prop_i : forall E Si Fi (p : Prop), p -> OP E Si Fi (<< p >>) true
  | op_prop_e : forall E Si Fi (p : Prop) q, OP E Si Fi (<< p >>) false ->
                                          (p -> OP E Si Fi q true) ->
                                          OP E Si Fi q true
  | op_and_i  : forall E Si Fi p q, OP E Si Fi p true -> OP E Si Fi q true ->
                                                   OP E Si Fi (p ./\ q) true
  | op_and_e1 : forall E Si Fi p q, OP E Si Fi (p ./\ q) false -> OP E Si Fi p false
  | op_and_e2 : forall E Si Fi p q, OP E Si Fi (p ./\ q) false -> OP E Si Fi q false
  | op_or_i1  : forall E Si Fi p q, OP E Si Fi p true -> OP E Si Fi (p .\/ q) true
  | op_or_i2  : forall E Si Fi p q, OP E Si Fi q true -> OP E Si Fi (p .\/ q) true
  | op_or_e   : forall E Si Fi p q r, OP E Si Fi (p .\/ q) false ->
                                   OP (cons p E) Si Fi r true ->
                                   OP (cons q E) Si Fi r true ->
                                   OP E Si Fi r true
  | op_imp_i  : forall E Si Fi p q, OP (cons p E) Si Fi q true ->
                                 OP E Si Fi (p .-> q) true
  | op_imp_e  : forall E Si Fi p q, OP E Si Fi p true -> OP E Si Fi (p .-> q) false ->
                                                   OP E Si Fi q false
  | op_allx_i : forall E Si Fi A (P : A -> _),
                   (forall x, OP E Si Fi (P x) true) -> OP E Si Fi (allx _ _ P) true
  | op_allx_e : forall E Si Fi A P, OP E Si Fi (allx _ _ P) false
                                            -> forall B : A, OP E Si Fi (P B) false
  | op_extx_i : forall E Si Fi A (P : A -> _) x,
                   OP E Si Fi (P x) true -> OP E Si Fi (extx _ _ P) true
  | op_extx_e : forall E Si Fi A P q,
                   OP E Si Fi (extx _ _ P) false ->
                   (forall B : A, OP (cons (P B) E) Si Fi q true) -> OP E Si Fi q true
  | op_allv_i : forall E Si Fi A p,
                  (forall x : A -> _, OP E Si Fi (Subst _ _ p x) true)
                                                   -> OP E Si Fi (allv tO _ p) true
  | op_extv_i : forall E Si Fi A p (q:A->_),
                  OP E Si Fi (Subst _ _ p q) true     -> OP E Si Fi (extv tO _ p) true
  | op_mu_i   : forall E Si Fi A p (v:A),
                  OP E Si Fi (Subst _ _ (p v) (fun v => mu tO _ p v)) true
                    -> OP E Si Fi (mu tO _ p v) true
.

Theorem Soundness_of_Annotated_Deductions :
  forall E Si Fi p b, OP E Si Fi p b -> OK E Si Fi p.
induction 1.
auto.
auto.
apply ok_env. auto.
apply ok_cptr_i. auto.
apply ok_cptr_e with l P. auto. auto.
apply ok_ref_i. auto.
apply ok_ref_e with l P. auto. auto.
apply ok_prop_i. auto.
apply ok_prop_e with p. auto. auto.
apply ok_and_i. auto. auto.
apply ok_and_e1 with q. auto. auto.
apply ok_and_e2 with p. auto. auto.
apply ok_or_i1. auto.
apply ok_or_i2. auto.
apply ok_or_e with p q. auto. auto. auto.
apply ok_imp_i. auto.
apply ok_imp_e with p. auto. auto.
apply ok_allx_i. auto.
apply ok_allx_e with A. auto.
apply ok_extx_i with x. auto.
apply ok_extx_e with A P. auto. auto.
apply ok_allv_i. auto.
apply ok_extv_i with q. auto.
apply ok_mu_i. auto.
Qed.

Theorem Completeness_of_Annotated_Deductions :
  forall E Si Fi p, OK E Si Fi p -> OP E Si Fi p true.
induction 1.
apply op_coer. apply op_env. auto.
apply op_cptr_i. auto.
apply op_cptr_e with l P. apply op_coer'. auto. auto.
apply op_ref_i. auto.
apply op_ref_e with l P. apply op_coer'. auto. auto.
apply op_prop_i. auto.
apply op_prop_e with p. apply op_coer'. auto. auto.
apply op_and_i. auto. auto.
apply op_coer. apply op_and_e1 with q. apply op_coer'. auto.
apply op_coer. apply op_and_e2 with p. apply op_coer'. auto.
apply op_or_i1. auto.
apply op_or_i2. auto.
apply op_or_e with p q. auto. apply op_coer'. auto. auto. auto.
apply op_imp_i. auto.
apply op_coer. apply op_imp_e with p. auto. apply op_coer'. auto.
apply op_allx_i. auto.
apply op_coer. apply op_allx_e with A. apply op_coer'. auto.
apply op_extx_i with x. auto.
apply op_extx_e with A P. apply op_coer'. auto. auto.
apply op_allv_i. auto.
apply op_extv_i with q. auto.
apply op_mu_i. auto.
Qed.

Definition wk E E' := forall p, InE E p -> InE E' p.

Lemma wk_cons_wk :
  forall E E' p, wk E E' -> wk (cons p E) (cons p E').
unfold wk. intros. destruct H0. rewrite H0. left. auto. right. auto.
Qed.

Lemma Weaken_Property_for_Extractions :
  forall E Si Fi p b, OB E Si Fi p b -> forall E', wk E E' -> OB E' Si Fi p b.
induction 1; intros.
apply ob_coer. auto.
apply ob_env. auto.
apply ob_cptr_i. auto.
apply ob_cptr_e with l P. auto. auto.
apply ob_ref_i. auto.
apply ob_ref_e with l P. auto. auto.
apply ob_prop_i. auto.
apply ob_prop_e with p. auto. auto.
apply ob_and_i. auto. auto.
apply ob_and_e1 with q. auto.
apply ob_and_e2 with p. auto.
apply ob_or_i1. auto.
apply ob_or_i2. auto.
apply ob_or_e with p q. auto. apply IHOB2. apply wk_cons_wk. auto. apply IHOB3.
 apply wk_cons_wk. auto.
apply ob_imp_i. apply IHOB. apply wk_cons_wk. auto.
apply ob_imp_e with p. apply IHOB1. auto. apply IHOB2. auto.
apply ob_allx_i. auto.
apply ob_allx_e with A. auto.
apply ob_extx_i with x. auto.
apply ob_extx_e with A P. auto. intro. apply H1 with B. apply wk_cons_wk. auto.
apply ob_allv_i. auto.
apply ob_extv_i with q. auto.
apply ob_mu_i. auto.
Qed.

Lemma Weaken_Property_for_Extractions_A :
  forall E Si Fi p b, OP E Si Fi p b -> forall E', wk E E' -> OP E' Si Fi p b.
induction 1; intros.
apply op_coer. auto.
apply op_coer'. auto.
apply op_env. auto.
apply op_cptr_i. auto.
apply op_cptr_e with l P. auto. auto.
apply op_ref_i. auto.
apply op_ref_e with l P. auto. auto.
apply op_prop_i. auto.
apply op_prop_e with p. auto. auto.
apply op_and_i. auto. auto.
apply op_and_e1 with q. auto.
apply op_and_e2 with p. auto.
apply op_or_i1. auto.
apply op_or_i2. auto.
apply op_or_e with p q. auto. apply IHOP2. apply wk_cons_wk. auto. apply IHOP3.
 apply wk_cons_wk. auto.
apply op_imp_i. apply IHOP. apply wk_cons_wk. auto.
apply op_imp_e with p. apply IHOP1. auto. apply IHOP2. auto.
apply op_allx_i. auto.
apply op_allx_e with A. auto.
apply op_extx_i with x. auto.
apply op_extx_e with A P. auto. intro. apply H1 with B. apply wk_cons_wk. auto.
apply op_allv_i. auto.
apply op_extv_i with q. auto.
apply op_mu_i. auto.
Qed.

Lemma cons_wk : forall E p, wk E (cons p E).
unfold wk. simpl. tauto.
Qed.

Lemma wk_permut_wk :
  forall E E' p q, wk E (cons p (cons q E')) -> wk E (cons q (cons p E')).
unfold wk. simpl. intros. generalize (H _ H0). tauto.
Qed.

Lemma Substitution_Property_for_Extractions :
  forall E Si Fi p b, OB E Si Fi p b ->
    forall E' q, OB E' Si Fi q false -> wk E (cons q E') -> OB E' Si Fi p b.
induction 1; intros.
apply ob_coer. apply IHOB with q. auto. auto.
destruct (H1 _ H). rewrite H2. auto. apply ob_env. auto.
apply ob_cptr_i. auto.
apply ob_cptr_e with l P. apply IHOB with q0. auto. auto. intro.
 apply H1 with q0. auto. auto. auto.
apply ob_ref_i. auto.
apply ob_ref_e with l P. apply IHOB with q0. auto. auto. intro.
 apply H1 with q0. auto. auto. auto.
apply ob_prop_i. auto.
apply ob_prop_e with p. apply IHOB with q0. auto. auto. intro. apply H1 with q0.
 auto. auto. auto.
apply ob_and_i. apply IHOB1 with q0. auto. auto. apply IHOB2 with q0. auto. auto.
apply ob_and_e1 with q. apply IHOB with q0. auto. auto.
apply ob_and_e2 with p. apply IHOB with q0. auto. auto.
apply ob_or_i1. apply IHOB with q0. auto. auto.
apply ob_or_i2. apply IHOB with q0. auto. auto.
apply ob_or_e with p q. apply IHOB1 with q0. auto. auto. apply IHOB2 with q0.
 apply Weaken_Property_for_Extractions with E'. auto. apply cons_wk.
 apply wk_permut_wk. apply wk_cons_wk. auto. apply IHOB3 with q0.
 apply Weaken_Property_for_Extractions with E'. auto. apply cons_wk.
 apply wk_permut_wk. apply wk_cons_wk. auto.
apply ob_imp_i. apply IHOB with q0.
 apply Weaken_Property_for_Extractions with E'. auto. apply cons_wk.
 apply wk_permut_wk. apply wk_cons_wk. auto.
apply ob_imp_e with p. apply IHOB1 with q0. auto. auto. apply IHOB2 with q0.
 auto. auto.
apply ob_allx_i. intro. apply H0 with q. auto. auto.
apply ob_allx_e with A. apply IHOB with q. auto. auto.
apply ob_extx_i with x. apply IHOB with q. auto. auto.
apply ob_extx_e with A P. apply IHOB with q0. auto. auto. intro.
 apply H1 with B q0. auto. apply Weaken_Property_for_Extractions with E'.
 auto. apply cons_wk. apply wk_permut_wk. apply wk_cons_wk. auto.
apply ob_allv_i. intro. apply H0 with q. auto. auto.
apply ob_extv_i with q. apply IHOB with q0. auto. auto.
apply ob_mu_i. apply IHOB with q. auto. auto.
Qed.

Lemma Substitution_Property_for_Extractions_A :
  forall E Si Fi p b, OP E Si Fi p b ->
    forall E' q, OP E' Si Fi q false -> wk E (cons q E') -> OP E' Si Fi p b.
induction 1; intros.
apply op_coer. apply IHOP with q. auto. auto.
apply op_coer'. apply IHOP with q. auto. auto.
destruct (H1 _ H). rewrite H2. auto. apply op_env. auto.
apply op_cptr_i. auto.
apply op_cptr_e with l P. apply IHOP with q0. auto. auto. intro.
 apply H1 with q0. auto. auto. auto.
apply op_ref_i. auto.
apply op_ref_e with l P. apply IHOP with q0. auto. auto. intro.
 apply H1 with q0. auto. auto. auto.
apply op_prop_i. auto.
apply op_prop_e with p. apply IHOP with q0. auto. auto. intro. apply H1 with q0.
 auto. auto. auto.
apply op_and_i. apply IHOP1 with q0. auto. auto. apply IHOP2 with q0. auto. auto.
apply op_and_e1 with q. apply IHOP with q0. auto. auto.
apply op_and_e2 with p. apply IHOP with q0. auto. auto.
apply op_or_i1. apply IHOP with q0. auto. auto.
apply op_or_i2. apply IHOP with q0. auto. auto.
apply op_or_e with p q. apply IHOP1 with q0. auto. auto. apply IHOP2 with q0.
 apply Weaken_Property_for_Extractions_A with E'. auto. apply cons_wk.
 apply wk_permut_wk. apply wk_cons_wk. auto. apply IHOP3 with q0.
 apply Weaken_Property_for_Extractions_A with E'. auto. apply cons_wk.
 apply wk_permut_wk. apply wk_cons_wk. auto.
apply op_imp_i. apply IHOP with q0.
 apply Weaken_Property_for_Extractions_A with E'. auto. apply cons_wk.
 apply wk_permut_wk. apply wk_cons_wk. auto.
apply op_imp_e with p. apply IHOP1 with q0. auto. auto. apply IHOP2 with q0.
 auto. auto.
apply op_allx_i. intro. apply H0 with q. auto. auto.
apply op_allx_e with A. apply IHOP with q. auto. auto.
apply op_extx_i with x. apply IHOP with q. auto. auto.
apply op_extx_e with A P. apply IHOP with q0. auto. auto. intro.
 apply H1 with B q0. auto. apply Weaken_Property_for_Extractions_A with E'.
 auto. apply cons_wk. apply wk_permut_wk. apply wk_cons_wk. auto.
apply op_allv_i. intro. apply H0 with q. auto. auto.
apply op_extv_i with q. apply IHOP with q0. auto. auto.
apply op_mu_i. apply IHOP with q. auto. auto.
Qed.

Inductive SEQ : Env -> CdHpSpec -> DtHpSpec -> PropX tO -> Prop :=
  | seq_env    : forall E Si Fi p, InE E p -> SEQ E Si Fi p
  | seq_cptr_r : forall E Si Fi l P, lookup Si l P -> SEQ E Si Fi (cptr _ l P)
  | seq_cptr_l : forall E Si Fi l P q, InE E (cptr _ l P) ->
                                     (lookup Si l P -> SEQ E Si Fi q) -> SEQ E Si Fi q
  | seq_ref_r : forall E Si Fi l P, lookup Fi l P -> SEQ E Si Fi (ref _ l P)
  | seq_ref_l : forall E Si Fi l P q, InE E (ref _ l P) ->
                                     (lookup Fi l P -> SEQ E Si Fi q) -> SEQ E Si Fi q
  | seq_prop_r : forall E Si Fi (p : Prop), p -> SEQ E Si Fi (<< p >>)
  | seq_prop_l : forall E Si Fi (p : Prop) q, InE E (<<p>>) -> (p -> SEQ E Si Fi q) ->
                                           SEQ E Si Fi q
  | seq_and_r  : forall E Si Fi p q, SEQ E Si Fi p -> SEQ E Si Fi q -> SEQ E Si Fi (p ./\ q)
  | seq_and_l1 : forall E Si Fi p q r, InE E (p ./\ q) -> SEQ (cons p E) Si Fi r ->
                                    SEQ E Si Fi r
  | seq_and_l2 : forall E Si Fi p q r, InE E (p ./\ q) -> SEQ (cons q E) Si Fi r ->
                                    SEQ E Si Fi r
  | seq_or_r1  : forall E Si Fi p q, SEQ E Si Fi p -> SEQ E Si Fi (p .\/ q)
  | seq_or_r2  : forall E Si Fi p q, SEQ E Si Fi q -> SEQ E Si Fi (p .\/ q)
  | seq_or_l   : forall E Si Fi p q r, InE E (p .\/ q) -> SEQ (cons p E) Si Fi r ->
                                    SEQ (cons q E) Si Fi r -> SEQ E Si Fi r
  | seq_imp_r  : forall E Si Fi p q, SEQ (cons p E) Si Fi q -> SEQ E Si Fi (p .-> q)
  | seq_imp_l  : forall E Si Fi p q r, InE E (p .-> q) -> SEQ E Si Fi p ->
                                    SEQ (cons q E) Si Fi r -> SEQ E Si Fi r
  | seq_allx_r : forall E Si Fi A (P : A -> _),
                  (forall x, SEQ E Si Fi (P x)) -> SEQ E Si Fi (allx _ _ P)
  | seq_allx_l : forall E Si Fi A P (B:A) q, InE E (allx _ _ P) ->
                                          SEQ (cons (P B) E) Si Fi q -> SEQ E Si Fi q
  | seq_extx_r : forall E Si Fi A (P : A -> _) x, SEQ E Si Fi (P x) ->
                                               SEQ E Si Fi (extx _ _ P)
  | seq_extx_l : forall E Si Fi A P q, InE E (extx _ _ P) ->
                                    (forall B:A, SEQ (cons (P B) E) Si Fi q) ->
                                    SEQ E Si Fi q
  | seq_allv_r : forall E Si Fi A p,
                  (forall x : A -> _, SEQ E Si Fi (Subst _ _ p x))
                                                   -> SEQ E Si Fi (allv tO _ p)
  | seq_extv_r : forall E Si Fi A p (q:A->_),
                     SEQ E Si Fi (Subst _ _ p q)      -> SEQ E Si Fi (extv tO _ p)
  | seq_mu_r   : forall E Si Fi A p (v:A),
                  SEQ E Si Fi (Subst _ _ (p v) (fun v => mu tO _ p v))
                    -> SEQ E Si Fi (mu tO _ p v)
.

Theorem Soundness_of_Sequent_Calculus :
  forall E Si Fi p, SEQ E Si Fi p -> OB E Si Fi p true.
intros. induction H.
apply ob_coer. apply ob_env. auto.
apply ob_cptr_i. auto.
apply ob_cptr_e with l P. apply ob_env. auto. auto.
apply ob_ref_i. auto.
apply ob_ref_e with l P. apply ob_env. auto. auto.
apply ob_prop_i. auto.
apply ob_prop_e with p. apply ob_env. auto. auto.
apply ob_and_i. auto. auto.
apply Substitution_Property_for_Extractions with (cons p E) p. auto.
 apply ob_and_e1 with q. apply ob_env. auto. unfold wk. auto.
apply Substitution_Property_for_Extractions with (cons q E) q. auto.
 apply ob_and_e2 with p. apply ob_env. auto. unfold wk. auto.
apply ob_or_i1. auto.
apply ob_or_i2. auto.
apply ob_or_e with p q. apply ob_env. auto. auto. auto.
apply ob_imp_i. auto.
apply Substitution_Property_for_Extractions with (cons q E) q. auto.
 apply ob_imp_e with p. auto. apply ob_env. auto. unfold wk. auto.
apply ob_allx_i. auto.
apply Substitution_Property_for_Extractions with (cons (P B) E) (P B). auto.
 apply ob_allx_e with A. apply ob_env. auto. unfold wk. auto.
apply ob_extx_i with x. auto.
apply ob_extx_e with A P. apply ob_env. auto. auto.
apply ob_allv_i. auto.
apply ob_extv_i with q. auto.
apply ob_mu_i. auto.
Qed.

Lemma Structural_Property_of_Sequent :
  forall E Si Fi p, SEQ E Si Fi p -> forall E', (forall p, InE E p -> InE E' p)
                                                                  -> SEQ E' Si Fi p.
induction 1; intros.
apply seq_env. auto.
apply seq_cptr_r. auto.
apply seq_cptr_l with l P. auto. auto.
apply seq_ref_r. auto.
apply seq_ref_l with l P. auto. auto.
apply seq_prop_r. auto.
apply seq_prop_l with p. auto. auto.
apply seq_and_r. auto. auto.
apply seq_and_l1 with p q. auto. apply IHSEQ. apply wk_cons_wk. auto.
apply seq_and_l2 with p q. auto. apply IHSEQ. apply wk_cons_wk. auto.
apply seq_or_r1. auto.
apply seq_or_r2. auto.
apply seq_or_l with p q. auto. apply IHSEQ1. apply wk_cons_wk. auto.
 apply IHSEQ2. apply wk_cons_wk. auto.
apply seq_imp_r. apply IHSEQ. apply wk_cons_wk. auto.
apply seq_imp_l with p q. auto. apply IHSEQ1. auto. apply IHSEQ2.
 apply wk_cons_wk. auto.
apply seq_allx_r. auto.
apply seq_allx_l with A P B. auto. apply IHSEQ. apply wk_cons_wk. auto.
apply seq_extx_r with x. auto.
apply seq_extx_l with A P. auto. intro. apply H1 with B.
 apply wk_cons_wk.
auto.
apply seq_allv_r. auto.
apply seq_extv_r with q. auto.
apply seq_mu_r. auto.
Qed.

Theorem Completeness_of_Sequent_Derivations :
  forall E Si Fi p b, OB E Si Fi p b ->
    if b then SEQ E Si Fi p else forall q, SEQ (cons p E) Si Fi q -> SEQ E Si Fi q.
induction 1; intros.
apply IHOB. apply seq_env. left. auto.
apply Structural_Property_of_Sequent with (cons p E). auto. intros. destruct H1.
 rewrite H1. auto. auto.
apply seq_cptr_r. auto.
apply IHOB. apply seq_cptr_l with l P. left. auto. intro.
 apply Structural_Property_of_Sequent with E. auto. intros. right. auto.
apply seq_ref_r. auto.
apply IHOB. apply seq_ref_l with l P. left. auto. intro.
 apply Structural_Property_of_Sequent with E. auto. intros. right. auto.
apply seq_prop_r. auto.
apply IHOB. apply seq_prop_l with p. left. auto. intro.
 apply Structural_Property_of_Sequent with E. auto. intros. right. auto.
apply seq_and_r. auto. auto.
apply IHOB. apply seq_and_l1 with p q. left. auto.
 apply Structural_Property_of_Sequent with (cons p E). auto. intros.
 destruct H1. left. auto. right. right. auto.
apply IHOB. apply seq_and_l2 with p q. left. auto.
 apply Structural_Property_of_Sequent with (cons q E). auto. intros.
 destruct H1. left. auto. right. right. auto.
apply seq_or_r1. auto.
apply seq_or_r2. auto.
apply IHOB1. apply seq_or_l with p q. left. auto.
 apply Structural_Property_of_Sequent with (cons p E). auto. intros.
 destruct H2. left. auto. right. right. auto.
 apply Structural_Property_of_Sequent with (cons q E). auto. intros.
 destruct H2. left. auto. right. right. auto.
apply seq_imp_r. auto.
apply IHOB2. apply seq_imp_l with p q. left. auto.
 apply Structural_Property_of_Sequent with E. auto. intros. right. auto.
 apply Structural_Property_of_Sequent with (cons q E). auto. intros.
 destruct H2. left. auto. right. right. auto.
apply seq_allx_r. auto.
apply IHOB. apply seq_allx_l with A P B. left. auto.
 apply Structural_Property_of_Sequent with (cons (P B) E). auto. intros.
 destruct H1. left. auto. right. right. auto.
apply seq_extx_r with x. auto.
apply IHOB. apply seq_extx_l with A P. left. auto. intro.
 apply Structural_Property_of_Sequent with (cons (P B) E). auto. intros.
 destruct H2. left. auto. right. right. auto.
apply seq_allv_r. auto.
apply seq_extv_r with q. auto.
apply seq_mu_r. auto.
Qed.

Inductive SEP : Env -> CdHpSpec -> DtHpSpec -> PropX tO -> Prop :=
  | sep_cut    : forall E Si Fi p q, SEP E Si Fi p -> SEP (cons p E) Si Fi q ->
                                                SEP E Si Fi q
  | sep_env    : forall E Si Fi p, InE E p -> SEP E Si Fi p
  | sep_cptr_r : forall E Si Fi l P, lookup Si l P -> SEP E Si Fi (cptr _ l P)
  | sep_cptr_l : forall E Si Fi l P q, InE E (cptr _ l P) ->
                                     (lookup Si l P -> SEP E Si Fi q) -> SEP E Si Fi q
  | sep_ref_r : forall E Si Fi l P, lookup Fi l P -> SEP E Si Fi (ref _ l P)
  | sep_ref_l : forall E Si Fi l P q, InE E (ref _ l P) ->
                                     (lookup Fi l P -> SEP E Si Fi q) -> SEP E Si Fi q
  | sep_prop_r : forall E Si Fi (p : Prop), p -> SEP E Si Fi (<< p >>)
  | sep_prop_l : forall E Si Fi (p : Prop) q, InE E (<<p>>) -> (p -> SEP E Si Fi q) ->
                                           SEP E Si Fi q
  | sep_and_r  : forall E Si Fi p q, SEP E Si Fi p -> SEP E Si Fi q -> SEP E Si Fi (p ./\ q)
  | sep_and_l1 : forall E Si Fi p q r, InE E (p ./\ q) -> SEP (cons p E) Si Fi r ->
                                    SEP E Si Fi r
  | sep_and_l2 : forall E Si Fi p q r, InE E (p ./\ q) -> SEP (cons q E) Si Fi r ->
                                    SEP E Si Fi r
  | sep_or_r1  : forall E Si Fi p q, SEP E Si Fi p -> SEP E Si Fi (p .\/ q)
  | sep_or_r2  : forall E Si Fi p q, SEP E Si Fi q -> SEP E Si Fi (p .\/ q)
  | sep_or_l   : forall E Si Fi p q r, InE E (p .\/ q) -> SEP (cons p E) Si Fi r ->
                                    SEP (cons q E) Si Fi r -> SEP E Si Fi r
  | sep_imp_r  : forall E Si Fi p q, SEP (cons p E) Si Fi q -> SEP E Si Fi (p .-> q)
  | sep_imp_l  : forall E Si Fi p q r, InE E (p .-> q) -> SEP E Si Fi p ->
                                    SEP (cons q E) Si Fi r -> SEP E Si Fi r
  | sep_allx_r : forall E Si Fi A (P : A -> _),
                  (forall x, SEP E Si Fi (P x)) -> SEP E Si Fi (allx _ _ P)
  | sep_allx_l : forall E Si Fi A P (B:A) q, InE E (allx _ _ P) ->
                                          SEP (cons (P B) E) Si Fi q -> SEP E Si Fi q
  | sep_extx_r : forall E Si Fi A (P : A -> _) x, SEP E Si Fi (P x) ->
                                               SEP E Si Fi (extx _ _ P)
  | sep_extx_l : forall E Si Fi A P q, InE E (extx _ _ P) ->
                                    (forall B:A, SEP (cons (P B) E) Si Fi q) ->
                                    SEP E Si Fi q
  | sep_allv_r : forall E Si Fi A p,
                  (forall x : A -> _, SEP E Si Fi (Subst _ _ p x))
                                                   -> SEP E Si Fi (allv tO _ p)
  | sep_extv_r : forall E Si Fi A p (q:A->_),
                     SEP E Si Fi (Subst _ _ p q)      -> SEP E Si Fi (extv tO _ p)
  | sep_mu_r   : forall E Si Fi A p (v:A),
                  SEP E Si Fi (Subst _ _ (p v) (fun v => mu tO _ p v))
                    -> SEP E Si Fi (mu tO _ p v)
.

Theorem Soundness_of_Sequent_Calculus_with_Cut :
  forall E Si Fi p, SEP E Si Fi p -> OP E Si Fi p true.
intros. induction H.
apply Substitution_Property_for_Extractions_A with (cons p E) p. auto.
 apply op_coer'. auto. unfold wk. auto.
apply op_coer. apply op_env. auto.
apply op_cptr_i. auto.
apply op_cptr_e with l P. apply op_env. auto. auto.
apply op_ref_i. auto.
apply op_ref_e with l P. apply op_env. auto. auto.
apply op_prop_i. auto.
apply op_prop_e with p. apply op_env. auto. auto.
apply op_and_i. auto. auto.
apply Substitution_Property_for_Extractions_A with (cons p E) p. auto.
 apply op_and_e1 with q. apply op_env. auto. unfold wk. auto.
apply Substitution_Property_for_Extractions_A with (cons q E) q. auto.
 apply op_and_e2 with p. apply op_env. auto. unfold wk. auto.
apply op_or_i1. auto.
apply op_or_i2. auto.
apply op_or_e with p q. apply op_env. auto. auto. auto.
apply op_imp_i. auto.
apply Substitution_Property_for_Extractions_A with (cons q E) q. auto.
 apply op_imp_e with p. auto. apply op_env. auto. unfold wk. auto.
apply op_allx_i. auto.
apply Substitution_Property_for_Extractions_A with (cons (P B) E) (P B). auto.
 apply op_allx_e with A. apply op_env. auto. unfold wk. auto.
apply op_extx_i with x. auto.
apply op_extx_e with A P. apply op_env. auto. auto.
apply op_allv_i. auto.
apply op_extv_i with q. auto.
apply op_mu_i. auto.
Qed.

Lemma Structural_Property_of_Sequent_with_Cut :
  forall E Si Fi p, SEP E Si Fi p -> forall E', (forall p, InE E p -> InE E' p)
                                                                  -> SEP E' Si Fi p.
induction 1; intros.
apply sep_cut with p. auto. apply IHSEP2. apply wk_cons_wk. auto.
apply sep_env. auto.
apply sep_cptr_r. auto.
apply sep_cptr_l with l P. auto. auto.
apply sep_ref_r. auto.
apply sep_ref_l with l P. auto. auto.
apply sep_prop_r. auto.
apply sep_prop_l with p. auto. auto.
apply sep_and_r. auto. auto.
apply sep_and_l1 with p q. auto. apply IHSEP. apply wk_cons_wk. auto.
apply sep_and_l2 with p q. auto. apply IHSEP. apply wk_cons_wk. auto.
apply sep_or_r1. auto.
apply sep_or_r2. auto.
apply sep_or_l with p q. auto. apply IHSEP1. apply wk_cons_wk. auto.
 apply IHSEP2. apply wk_cons_wk. auto.
apply sep_imp_r. apply IHSEP. apply wk_cons_wk. auto.
apply sep_imp_l with p q. auto. apply IHSEP1. auto. apply IHSEP2.
 apply wk_cons_wk. auto.
apply sep_allx_r. auto.
apply sep_allx_l with A P B. auto. apply IHSEP. apply wk_cons_wk. auto.
apply sep_extx_r with x. auto.
apply sep_extx_l with A P. auto. intro. apply H1 with B.
 apply wk_cons_wk. auto.
apply sep_allv_r. auto.
apply sep_extv_r with q. auto.
apply sep_mu_r. auto.
Qed.

Theorem Completeness_of_Sequent_Derivations_with_Cut :
  forall E Si Fi p b, OP E Si Fi p b ->
    if b then SEP E Si Fi p else forall q, SEP (cons p E) Si Fi q -> SEP E Si Fi q.
induction 1; intros.
apply IHOP. apply sep_env. left. auto.
apply sep_cut with p. auto. auto.
apply Structural_Property_of_Sequent_with_Cut with (cons p E). auto. intros.
 destruct H1. rewrite H1. auto. auto.
apply sep_cptr_r. auto.
apply IHOP. apply sep_cptr_l with l P. left. auto. intro.
 apply Structural_Property_of_Sequent_with_Cut with E. auto. intros. right.
 auto.
apply sep_ref_r. auto.
apply IHOP. apply sep_ref_l with l P. left. auto. intro.
 apply Structural_Property_of_Sequent_with_Cut with E. auto. intros. right.
 auto.
apply sep_prop_r. auto.
apply IHOP. apply sep_prop_l with p. left. auto. intro.
 apply Structural_Property_of_Sequent_with_Cut with E. auto. intros. right.
 auto.
apply sep_and_r. auto. auto.
apply IHOP. apply sep_and_l1 with p q. left. auto.
 apply Structural_Property_of_Sequent_with_Cut with (cons p E). auto. intros.
 destruct H1. left. auto. right. right. auto.
apply IHOP. apply sep_and_l2 with p q. left. auto.
 apply Structural_Property_of_Sequent_with_Cut with (cons q E). auto. intros.
 destruct H1. left. auto. right. right. auto.
apply sep_or_r1. auto.
apply sep_or_r2. auto.
apply IHOP1. apply sep_or_l with p q. left. auto.
 apply Structural_Property_of_Sequent_with_Cut with (cons p E). auto. intros.
 destruct H2. left. auto. right. right. auto.
 apply Structural_Property_of_Sequent_with_Cut with (cons q E). auto. intros.
 destruct H2. left. auto. right. right. auto.
apply sep_imp_r. auto.
apply IHOP2. apply sep_imp_l with p q. left. auto.
 apply Structural_Property_of_Sequent_with_Cut with E. auto. intros. right.
 auto. apply Structural_Property_of_Sequent_with_Cut with (cons q E). auto.
 intros. destruct H2. left. auto. right. right. auto.
apply sep_allx_r. auto.
apply IHOP. apply sep_allx_l with A P B. left. auto.
 apply Structural_Property_of_Sequent_with_Cut with (cons (P B) E). auto.
 intros. destruct H1. left. auto. right. right. auto.
apply sep_extx_r with x. auto.
apply IHOP. apply sep_extx_l with A P. left. auto. intro.
 apply Structural_Property_of_Sequent_with_Cut with (cons (P B) E). auto.
 intros. destruct H2. left. auto. right. right. auto.
apply sep_allv_r. auto.
apply sep_extv_r with q. auto.
apply sep_mu_r. auto.
Qed.

Lemma in_neq : forall E p E' q, InE E p -> wk E (cons q E') -> ~p=q -> InE E' p.
intros. destruct (H0 p). auto. tauto. auto.
Qed.

Lemma seq_cons : forall E E' Si Fi p q p', (forall q, InE E q -> InE (cons p' E') q) ->
                   ((forall q, InE (cons p E) q -> InE (cons p' (cons p E')) q) -> SEQ (cons p E') Si Fi q)
                   -> SEQ (cons p E') Si Fi q.
intros. apply H0. intros. destruct H1. rewrite H1. right. left. auto.
destruct (H _ H1). rewrite H2. left. auto. right. right. auto.
Qed.

Lemma seq_cons_wk : forall E E' Si Fi p p' q, ((forall q, InE E q -> InE (cons p (cons p' E')) q) -> SEQ (cons p' E') Si Fi q)
                         -> (forall q, InE E q -> InE (cons p E') q)
                         -> SEQ (cons p' E') Si Fi q.
intros. apply H. apply wk_permut_wk. unfold wk. intros. right. auto.
Qed.

Lemma seq_aux : forall E E' Si Fi p q r p', SEQ E Si Fi p -> SEQ E Si Fi q ->
                  (forall E, SEQ E Si Fi p -> SEQ E Si Fi q -> (forall q', InE (cons p' E') q' -> InE (cons (p ./\ q) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (p ./\ q) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H1. apply Structural_Property_of_Sequent with E. auto. right.
auto. apply Structural_Property_of_Sequent with E. auto. right. auto.
apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux1 : forall E E' Si Fi p q r p', SEQ E Si Fi p ->
                  (forall E, SEQ E Si Fi p -> (forall q', InE (cons p' E') q' -> InE (cons (p .\/ q) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (p .\/ q) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. apply Structural_Property_of_Sequent with E. auto. right.
auto.
apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux2 : forall E E' Si Fi p q r p', SEQ E Si Fi q ->
                  (forall E, SEQ E Si Fi q -> (forall q', InE (cons p' E') q' -> InE (cons (p .\/ q) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (p .\/ q) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. apply Structural_Property_of_Sequent with E. auto. right.
auto.
apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux3 : forall E E' Si Fi p q r p', SEQ (cons p E) Si Fi q ->
                  (forall E, SEQ (cons p E) Si Fi q -> (forall q', InE (cons p' E') q' -> InE (cons (p .-> q) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (p .-> q) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. apply Structural_Property_of_Sequent with (cons p E). auto.
intros. destruct H2. rewrite H2. left. auto. right. right. auto.
apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux4 : forall E E' Si Fi A q r p', (forall x, SEQ E Si Fi (q x)) ->
                  (forall E, (forall x, SEQ E Si Fi (q x)) -> (forall q', InE (cons p' E') q' -> InE (cons (allx _ A q) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (allx _ A q) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. intro. apply Structural_Property_of_Sequent with E. auto.
right. auto. apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux5 : forall E E' Si Fi A q r p', (exists x, SEQ E Si Fi (q x)) ->
                  (forall E, (exists x, SEQ E Si Fi (q x)) -> (forall q', InE (cons p' E') q' -> InE (cons (extx _ A q) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (extx _ A q) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. destruct H. exists x.
apply Structural_Property_of_Sequent with E. auto.
right. auto. apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux6 : forall E E' Si Fi A q r p', (forall x, SEQ E Si Fi (Subst _ _ q x)) ->
                  (forall E, (forall x, SEQ E Si Fi (Subst _ _ q x)) -> (forall q', InE (cons p' E') q' -> InE (cons (allv _ A (eq_rect _ _ q _ (eq_tplus_tO _))) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (allv _ A (eq_rect _ _ q _ (eq_tplus_tO _))) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. intro. apply Structural_Property_of_Sequent with E. auto.
right. auto. apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux7 : forall E E' Si Fi A q r p', (exists x, SEQ E Si Fi (Subst _ _ q x)) ->
                  (forall E, (exists x, SEQ E Si Fi (Subst _ _ q x)) -> (forall q', InE (cons p' E') q' -> InE (cons (extv _ A (eq_rect _ _ q _ (eq_tplus_tO _))) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (extv _ A (eq_rect _ _ q _ (eq_tplus_tO _))) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. destruct H. exists x.
apply Structural_Property_of_Sequent with E. auto.
right. auto. apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma seq_aux8 : forall E E' Si Fi A p v r p', SEQ E Si Fi (Subst _ _ (p v) (fun v => mu _ _ (fun v => (eq_rect _ _ (p v) _ (eq_tplus_tO _))) v)) ->
                  (forall E, SEQ E Si Fi (Subst _ _ (p v) (fun v => mu _ _ (fun v => (eq_rect _ _ (p v) _ (eq_tplus_tO _))) v)) ->
                    (forall q', InE (cons p' E') q' -> InE (cons (mu tO A p v) E) q') -> SEQ E Si Fi r)
                  -> (forall q', InE E' q' -> InE (cons (mu tO A p v) E) q')
                  -> SEQ (cons p' E) Si Fi r.
intros. apply H0. intros. apply Structural_Property_of_Sequent with E. auto.
right. auto. apply wk_permut_wk. apply wk_cons_wk. auto.
Qed.

Lemma Admissibility_of_Cut_cptr :
  forall w p Si Fi E p', SEQ E Si Fi p' -> forall E' q, SEQ E' Si Fi q -> p'=cptr _ w p
    -> (forall q, InE E' q -> InE (cons (cptr _ w p) E) q) -> SEQ E Si Fi q.
induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 rewrite <- H1 in H2. clear H1. generalize E H2. clear E H2. induction H0; intros.
  destruct (H2 _ H0). rewrite H1. apply seq_cptr_r. auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  destruct (H3 _ H0). apply H2. inversion H4. rewrite (inj_pairT2 _ _ _ _ _ H7).
    auto. auto. auto. apply seq_cptr_l with l0 P0. auto. intro. auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l0 P0. apply in_neq with E (cptr _ l P); auto. discriminate.
    intro. apply H2; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p0. apply in_neq with E (cptr _ l P); auto. discriminate.
    intro. apply H2; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (cptr _ l P); auto.
    discriminate. apply seq_cons with E (cptr _ l P); auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (cptr _ l P); auto.
    discriminate. apply seq_cons with E (cptr _ l P); auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (cptr _ l P); auto.
    discriminate. apply seq_cons with E (cptr _ l P); auto.
    apply seq_cons with E (cptr _ l P); auto.
  apply seq_imp_r. apply seq_cons with E (cptr _ l P); auto.
  apply seq_imp_l with p0 q. apply in_neq with E (cptr _ l P); auto.
    discriminate. apply IHSEQ1; auto. apply seq_cons with E (cptr _ l P); auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A P0 B. apply in_neq with E (cptr _ l P); auto.
    discriminate. apply seq_cons with E (cptr _ l P); auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A P0. apply in_neq with E (cptr _ l P); auto.
    discriminate. intro. apply seq_cons with E (cptr _ l P); auto. apply H2. auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p0; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p0 q; auto. apply seq_cons_wk with E' (cptr _ w p); eauto.
 apply seq_and_l2 with p0 q; auto. apply seq_cons_wk with E' (cptr _ w p); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p0 q; auto. apply seq_cons_wk with E' (cptr _ w p); eauto.
   apply seq_cons_wk with E' (cptr _ w p); eauto.
 inversion H1.
 apply seq_imp_l with p0 q; auto. apply seq_cons_wk with E' (cptr _ w p); eauto.
 inversion H2.
 apply seq_allx_l with A P B; auto. apply seq_cons_wk with E' (cptr _ w p); eauto.
 inversion H1.
 apply seq_extx_l with A P; auto. intro. apply seq_cons_wk with E' (cptr _ w p);
   eauto.
 inversion H2.
 inversion H1.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_ref :
  forall w p Si Fi E p', SEQ E Si Fi p' -> forall E' q, SEQ E' Si Fi q -> p'=ref _ w p
    -> (forall q, InE E' q -> InE (cons (ref _ w p) E) q) -> SEQ E Si Fi q.
induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 rewrite <- H1 in H2. clear H1. generalize E H2. clear E H2. induction H0; intros.
  destruct (H2 _ H0). rewrite H1. apply seq_ref_r. auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l0 P0. apply in_neq with E (ref _ l P); auto. discriminate.
    intro. apply H2; auto.
  apply seq_ref_r. auto.
  destruct (H3 _ H0). apply H2. inversion H4. rewrite (inj_pairT2 _ _ _ _ _ H7).
    auto. auto. auto. apply seq_ref_l with l0 P0. auto. intro. auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p0. apply in_neq with E (ref _ l P); auto. discriminate.
    intro. apply H2; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (ref _ l P); auto.
    discriminate. apply seq_cons with E (ref _ l P); auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (ref _ l P); auto.
    discriminate. apply seq_cons with E (ref _ l P); auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (ref _ l P); auto.
    discriminate. apply seq_cons with E (ref _ l P); auto.
    apply seq_cons with E (ref _ l P); auto.
  apply seq_imp_r. apply seq_cons with E (ref _ l P); auto.
  apply seq_imp_l with p0 q. apply in_neq with E (ref _ l P); auto.
    discriminate. apply IHSEQ1; auto. apply seq_cons with E (ref _ l P); auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A P0 B. apply in_neq with E (ref _ l P); auto.
    discriminate. apply seq_cons with E (ref _ l P); auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A P0. apply in_neq with E (ref _ l P); auto.
    discriminate. intro. apply seq_cons with E (ref _ l P); auto. apply H2. auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p0; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p0 q; auto. apply seq_cons_wk with E' (ref _ w p); eauto.
 apply seq_and_l2 with p0 q; auto. apply seq_cons_wk with E' (ref _ w p); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p0 q; auto. apply seq_cons_wk with E' (ref _ w p); eauto.
   apply seq_cons_wk with E' (ref _ w p); eauto.
 inversion H1.
 apply seq_imp_l with p0 q; auto. apply seq_cons_wk with E' (ref _ w p); eauto.
 inversion H2.
 apply seq_allx_l with A P B; auto. apply seq_cons_wk with E' (ref _ w p); eauto.
 inversion H1.
 apply seq_extx_l with A P; auto. intro. apply seq_cons_wk with E' (ref _ w p);
   eauto.
 inversion H2.
 inversion H1.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_prop :
  forall P Si Fi E p', SEQ E Si Fi p' -> forall E' q, SEQ E' Si Fi q -> p'=<<P>> ->
    (forall q, InE E' q -> InE (cons (<<P>>) E) q) -> SEQ E Si Fi q.
 induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P0; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P0; auto. intro. apply H1 with E'; auto.
 rewrite <- H1 in H2. clear H1. generalize E H2. clear E H2. induction H0; intros.
  destruct (H2 _ H0). rewrite H1. apply seq_prop_r. auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P0. apply in_neq with E (prop tO p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P0. apply in_neq with E (prop tO p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_prop_r. auto.
  destruct (H3 _ H0). apply H2. inversion H4. auto. auto. auto.
    apply seq_prop_l with p0. auto. intro. auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (prop tO p); auto.
    discriminate. apply seq_cons with E (prop tO p); auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (prop tO p); auto.
    discriminate. apply seq_cons with E (prop tO p); auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (prop tO p); auto.
    discriminate. apply seq_cons with E (prop tO p); auto.
    apply seq_cons with E (prop tO p); auto.
  apply seq_imp_r. apply seq_cons with E (prop tO p); auto.
  apply seq_imp_l with p0 q. apply in_neq with E (prop tO p); auto.
    discriminate. apply IHSEQ1; auto. apply seq_cons with E (prop tO p); auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A P0 B. apply in_neq with E (prop tO p); auto.
    discriminate. apply seq_cons with E (prop tO p); auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A P0. apply in_neq with E (prop tO p); auto.
    discriminate. intro. apply seq_cons with E (prop tO p); auto. apply H2. auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 apply seq_prop_l with p; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p q; auto. apply seq_cons_wk with E' (prop tO P); eauto.
 apply seq_and_l2 with p q; auto. apply seq_cons_wk with E' (prop tO P); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p q; auto. apply seq_cons_wk with E' (prop tO P); eauto.
   apply seq_cons_wk with E' (prop tO P); eauto.
 inversion H1.
 apply seq_imp_l with p q; auto. apply seq_cons_wk with E' (prop tO P); eauto.
 inversion H2.
 apply seq_allx_l with A P0 B; auto. apply seq_cons_wk with E' (prop tO P); eauto.
 inversion H1.
 apply seq_extx_l with A P0; auto. intro. apply seq_cons_wk with E' (prop tO P);
   eauto.
 inversion H2.
 inversion H1.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_and :
forall p1 (IHp1 : forall Si Fi (E : Env) p',
         SEQ E Si Fi p' ->
         forall (E' : Env) q,
         SEQ E' Si Fi q ->
         p' = p1 ->
         (forall q, InE E' q -> InE (cons p1 E) q) -> SEQ E Si Fi q)
  p2  (IHp2 : forall Si Fi (E : Env) p',
         SEQ E Si Fi p' ->
         forall (E' : Env) q,
         SEQ E' Si Fi q ->
         p' = p2 ->
         (forall q, InE E' q -> InE (cons p2 E) q) -> SEQ E Si Fi q),
   forall Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall (E' : Env) q,
   SEQ E' Si Fi q ->
   p' = p1 ./\ p2 ->
   (forall q, InE E' q -> InE (cons (p1 ./\ p2) E) q) -> SEQ E Si Fi q.
 intros p1 IHp1 p2 IHp2. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p; auto. intro. apply H1 with E'; auto.
 inversion H2. generalize (inj_pairT2 _ _ _ _ _ H6).
   generalize (inj_pairT2 _ _ _ _ _ H5). clear H5 H6. intros H5 H6.
   rewrite H5 in H. rewrite H5 in IHSEQ1. rewrite H6 in H0.
   rewrite H6 in IHSEQ2. clear p q H2 H5 H6.
   generalize E H H0 H3. clear E H H0 H3 IHSEQ1 IHSEQ2. induction H1; intros.
  destruct (H3 _ H). rewrite H2. apply seq_and_r; auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (p1 ./\ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (p1 ./\ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p. apply in_neq with E (p1 ./\ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_and_r; auto.
  destruct (H3 _ H). inversion H4. apply IHp1 with p1 (cons p E0); auto.
    apply seq_aux with E p1 p2; auto. rewrite (inj_pairT2 _ _ _ _ _ H6). auto.
    apply seq_and_l1 with p q; auto. apply seq_aux with E p1 p2; auto.
  destruct (H3 _ H). inversion H4. apply IHp2 with p2 (cons q E0); auto.
    apply seq_aux with E p1 p2; auto. rewrite (inj_pairT2 _ _ _ _ _ H7). auto.
    apply seq_and_l2 with p q; auto. apply seq_aux with E p1 p2; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p q. apply in_neq with E (p1 ./\ p2); auto. discriminate.
    apply seq_aux with E p1 p2; auto. apply seq_aux with E p1 p2; auto.
  apply seq_imp_r. apply seq_aux with E p1 p2; auto.
  apply seq_imp_l with p q. apply in_neq with E (p1 ./\ p2); auto. discriminate.
    auto. apply seq_aux with E p1 p2; auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A P B. apply in_neq with E (p1 ./\ p2); auto.
    discriminate. apply seq_aux with E p1 p2; auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A P. apply in_neq with E (p1 ./\ p2); auto.
    discriminate. intro. apply seq_aux with E p1 p2; auto. intros.
    apply H1 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 apply seq_and_l1 with p q; auto. apply seq_cons_wk with E' (p1 ./\ p2); eauto.
 apply seq_and_l2 with p q; auto. apply seq_cons_wk with E' (p1 ./\ p2); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p q; auto. apply seq_cons_wk with E' (p1 ./\ p2); eauto.
   apply seq_cons_wk with E' (p1 ./\ p2); eauto.
 inversion H1.
 apply seq_imp_l with p q; auto. apply seq_cons_wk with E' (p1 ./\ p2); eauto.
 inversion H2.
 apply seq_allx_l with A P B; auto. apply seq_cons_wk with E' (p1 ./\ p2); eauto.
 inversion H1.
 apply seq_extx_l with A P; auto. intro. apply seq_cons_wk with E' (p1 ./\ p2);
   eauto.
 inversion H2.
 inversion H1.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_or :
  forall p1,
  (forall Si Fi (E : Env) p',
         SEQ E Si Fi p' ->
         forall (E' : Env) q,
         SEQ E' Si Fi q ->
         p' = p1 ->
         (forall q, InE E' q -> InE (cons p1 E) q) -> SEQ E Si Fi q) ->
  forall p2,
  (forall Si Fi (E : Env) p',
         SEQ E Si Fi p' ->
         forall (E' : Env) q,
         SEQ E' Si Fi q ->
         p' = p2 ->
         (forall q, InE E' q -> InE (cons p2 E) q) -> SEQ E Si Fi q) ->
   forall Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall (E' : Env) q,
   SEQ E' Si Fi q ->
   p' = (p1 .\/ p2) ->
   (forall q, InE E' q -> InE (cons (p1 .\/ p2) E) q) -> SEQ E Si Fi q.
 intros p1 IHp1 p2 IHp2. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p q; auto. apply seq_cons_wk with E' (p1 .\/ p2); eauto.
 apply seq_and_l2 with p q; auto. apply seq_cons_wk with E' (p1 .\/ p2); eauto.
 inversion H1. rewrite (inj_pairT2 _ _ _ _ _ H4) in H.
   rewrite (inj_pairT2 _ _ _ _ _ H4) in IHSEQ. clear p q H1 H4 H5.
   generalize E H H2. clear E H H2 IHSEQ. induction H0; intros.
  destruct (H2 _ H). rewrite H1. apply seq_or_r1; auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p q. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    apply seq_aux1 with E p1 p2; auto.
  apply seq_and_l2 with p q. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    apply seq_aux1 with E p1 p2; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  destruct (H2 _ H). inversion H1. apply IHp1 with p1 (cons p E0); auto.
    apply seq_aux1 with E p1 p2; auto. rewrite (inj_pairT2 _ _ _ _ _ H4). auto.
    apply seq_or_l with p q; auto. apply seq_aux1 with E p1 p2; auto.
    apply seq_aux1 with E p1 p2; auto.
  apply seq_imp_r. apply seq_aux1 with E p1 p2; auto.
  apply seq_imp_l with p q. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    auto. apply seq_aux1 with E p1 p2; auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A P B. apply in_neq with E (p1 .\/ p2); auto.
    discriminate. apply seq_aux1 with E p1 p2; auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A P. apply in_neq with E (p1 .\/ p2); auto.
    discriminate. intro. apply seq_aux1 with E p1 p2; auto. intros.
    apply H1 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 inversion H1. rewrite (inj_pairT2 _ _ _ _ _ H5) in H.
   rewrite (inj_pairT2 _ _ _ _ _ H5) in IHSEQ. clear p q H1 H4 H5.
   generalize E H H2. clear E H H2 IHSEQ. induction H0; intros.
  destruct (H2 _ H). rewrite H1. apply seq_or_r2; auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p q. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    apply seq_aux2 with E p1 p2; auto.
  apply seq_and_l2 with p q. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    apply seq_aux2 with E p1 p2; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  destruct (H2 _ H). inversion H1. apply IHp2 with p2 (cons q E0); auto.
    apply seq_aux2 with E p1 p2; auto. rewrite (inj_pairT2 _ _ _ _ _ H5). auto.
    apply seq_or_l with p q; auto. apply seq_aux2 with E p1 p2; auto.
    apply seq_aux2 with E p1 p2; auto.
  apply seq_imp_r. apply seq_aux2 with E p1 p2; auto.
  apply seq_imp_l with p q. apply in_neq with E (p1 .\/ p2); auto. discriminate.
    auto. apply seq_aux2 with E p1 p2; auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A P B. apply in_neq with E (p1 .\/ p2); auto.
    discriminate. apply seq_aux2 with E p1 p2; auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A P. apply in_neq with E (p1 .\/ p2); auto.
    discriminate. intro. apply seq_aux2 with E p1 p2; auto. intros.
    apply H1 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 apply seq_or_l with p q; auto. apply seq_cons_wk with E' (p1 .\/ p2); eauto.
   apply seq_cons_wk with E' (p1 .\/ p2); eauto.
 inversion H1.
 apply seq_imp_l with p q; auto. apply seq_cons_wk with E' (p1 .\/ p2); eauto.
 inversion H2.
 apply seq_allx_l with A P B; auto. apply seq_cons_wk with E' (p1 .\/ p2); eauto.
 inversion H1.
 apply seq_extx_l with A P; auto. intro. apply seq_cons_wk with E' (p1 .\/ p2);
   eauto.
 inversion H2.
 inversion H1.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_imp :
  forall p1,
  (forall Si Fi (E : Env) p',
         SEQ E Si Fi p' ->
         forall (E' : Env) q,
         SEQ E' Si Fi q ->
         p' = p1 ->
         (forall q, InE E' q -> InE (cons p1 E) q) -> SEQ E Si Fi q) ->
  forall p2,
  (forall Si Fi (E : Env) p',
         SEQ E Si Fi p' ->
         forall (E' : Env) q,
         SEQ E' Si Fi q ->
         p' = p2 ->
         (forall q, InE E' q -> InE (cons p2 E) q) -> SEQ E Si Fi q) ->
   forall Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall (E' : Env) q,
   SEQ E' Si Fi q ->
   p' = (p1 .-> p2) ->
   (forall q, InE E' q -> InE (cons (p1 .-> p2) E) q) -> SEQ E Si Fi q.
 intros p1 IHp1 p2 IHp2. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p q; auto. apply seq_cons_wk with E' (p1 .-> p2); eauto.
 apply seq_and_l2 with p q; auto. apply seq_cons_wk with E' (p1 .-> p2); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p q; auto. apply seq_cons_wk with E' (p1 .-> p2); eauto.
   apply seq_cons_wk with E' (p1 .-> p2); eauto.
 inversion H1. rewrite (inj_pairT2 _ _ _ _ _ H4) in H.
   rewrite (inj_pairT2 _ _ _ _ _ H4) in IHSEQ.
   rewrite (inj_pairT2 _ _ _ _ _ H5) in H.
   rewrite (inj_pairT2 _ _ _ _ _ H5) in IHSEQ. clear p q H1 H4 H5.
   generalize E H H2. clear E H H2 IHSEQ. induction H0; intros.
  destruct (H2 _ H). rewrite H1. apply seq_imp_r; auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (p1 .-> p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (p1 .-> p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p. apply in_neq with E (p1 .-> p2); auto. discriminate.
    intro. apply H1; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p q. apply in_neq with E (p1 .-> p2); auto. discriminate.
    apply seq_aux3 with E p1 p2; auto.
  apply seq_and_l2 with p q. apply in_neq with E (p1 .-> p2); auto. discriminate.
    apply seq_aux3 with E p1 p2; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p q. apply in_neq with E (p1 .-> p2); auto. discriminate.
    apply seq_aux3 with E p1 p2; auto. apply seq_aux3 with E p1 p2; auto.
  apply seq_imp_r. apply seq_aux3 with E p1 p2; auto.
  destruct (H2 _ H). inversion H1. rewrite (inj_pairT2 _ _ _ _ _ H4) in H.
    rewrite (inj_pairT2 _ _ _ _ _ H4) in H0_.
    rewrite (inj_pairT2 _ _ _ _ _ H4) in IHSEQ1.
    rewrite (inj_pairT2 _ _ _ _ _ H5) in H.
    rewrite (inj_pairT2 _ _ _ _ _ H5) in H0_0.
    rewrite (inj_pairT2 _ _ _ _ _ H5) in IHSEQ2. clear p q H1 H4 H5.
    apply IHp2 with p2 (cons p2 E0); auto.
    apply IHp1 with p1 (cons p1 E0); auto.
    apply IHSEQ2; auto. apply Structural_Property_of_Sequent with (cons p1 E0).
    auto. intros. destruct H1. rewrite H1. left. auto. right. right. auto.
    apply wk_permut_wk. apply wk_cons_wk. auto.
    apply seq_imp_l with p q; auto. apply seq_aux3 with E p1 p2; auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A P B. apply in_neq with E (p1 .-> p2); auto.
    discriminate. apply seq_aux3 with E p1 p2; auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A P. apply in_neq with E (p1 .-> p2); auto.
    discriminate. intro. apply seq_aux3 with E p1 p2; auto. intros.
    apply H1 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 apply seq_imp_l with p q; auto. apply seq_cons_wk with E' (p1 .-> p2); eauto.
 inversion H2.
 apply seq_allx_l with A P B; auto. apply seq_cons_wk with E' (p1 .-> p2); eauto.
 inversion H1.
 apply seq_extx_l with A P; auto. intro. apply seq_cons_wk with E' (p1 .-> p2);
   eauto.
 inversion H2.
 inversion H1.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_allx :
  forall (A : Type) (p : A -> _),
     (forall (a : A) Si Fi (E : Env) p',
      SEQ E Si Fi p' ->
      forall (E' : Env) q,
      SEQ E' Si Fi q ->
      p' = p a ->
      (forall q, InE E' q -> InE (cons (p a) E) q) -> SEQ E Si Fi q) ->
   forall Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall (E' : Env) q,
   SEQ E' Si Fi q ->
   p' = allx _ A p ->
   (forall q, InE E' q -> InE (cons (allx _ A p) E) q) -> SEQ E Si Fi q.
intros A p H. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H3 _ H4). rewrite H5. rewrite H2 in H0. auto. auto.
 inversion H2.
 apply seq_cptr_l with l P; auto. intro. apply H2 with E'; auto.
 inversion H2.
 apply seq_ref_l with l P; auto. intro. apply H2 with E'; auto.
 inversion H2.
 apply seq_prop_l with p0; auto. intro. apply H2 with E'; auto.
 inversion H1.
 apply seq_and_l1 with p0 q; auto. apply seq_cons_wk with E' (allx _ A p); eauto.
 apply seq_and_l2 with p0 q; auto. apply seq_cons_wk with E' (allx _ A p); eauto.
 inversion H2.
 inversion H2.
 apply seq_or_l with p0 q; auto. apply seq_cons_wk with E' (allx _ A p); eauto.
   apply seq_cons_wk with E' (allx _ A p); eauto.
 inversion H2.
 apply seq_imp_l with p0 q; auto. apply seq_cons_wk with E' (allx _ A p); eauto.
 generalize H4. clear H1. cut (forall x, SEQ E Si Fi (p x)). generalize E.
   clear E A0 P H0 H3 H4. induction H2; intros.
  destruct (H4 _ H0). rewrite H2. apply seq_allx_r; auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (allx _ A p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (allx _ A p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p0. apply in_neq with E (allx _ A p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (allx _ A p); auto. discriminate.
    apply seq_aux4 with E A p; auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (allx _ A p); auto. discriminate.
    apply seq_aux4 with E A p; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (allx _ A p); auto. discriminate.
    apply seq_aux4 with E A p; auto. apply seq_aux4 with E A p; auto.
  apply seq_imp_r. apply seq_aux4 with E A p; auto.
  apply seq_imp_l with p0 q. apply in_neq with E (allx _ A p); auto. discriminate.
    auto. apply seq_aux4 with E A p; auto.
  apply seq_allx_r. auto.
  destruct (H4 _ H0). rewrite H3 in H0.
    cut (exists B, SEQ (cons (p B) E) Si Fi q
                /\ (forall E0, (forall x, SEQ E0 Si Fi (p x)) -> (forall q, InE (cons (p B) E) q -> InE (cons (allx _ A p) E0) q) -> SEQ E0 Si Fi q)).
    clear A0 P B H2 H3 IHSEQ. intro. destruct H2 as [B]. destruct H2.
    apply H with B (p B) (cons (p B) E0); auto.
    apply seq_aux4 with E A p; auto.
    generalize IHSEQ. inversion H3. generalize (inj_pairT2 _ _ _ _ _ H7). intro.
    inversion H5. intros. exists B. auto.
  apply seq_allx_l with A0 P B. auto. apply seq_aux4 with E A p; auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A0 P. apply in_neq with E (allx _ A p); auto.
    discriminate. intro. apply seq_aux4 with E A p; auto. intros.
    apply H2 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
  inversion H3. generalize (inj_pairT2 _ _ _ _ _ H6). intro. inversion H1.
    intro. apply H0.
 apply seq_allx_l with A0 P B; auto. apply seq_cons_wk with E' (allx _ A p); eauto.
 inversion H2.
 apply seq_extx_l with A0 P; auto. intro. apply seq_cons_wk with E' (allx _ A p);
   eauto.
 inversion H3.
 inversion H2.
 inversion H2.
Qed.

Lemma Admissibility_of_Cut_extx :
  forall (A : Type) (p : A -> _),
   (forall (a : A) Si Fi (E : Env) p',
      SEQ E Si Fi p' ->
      forall (E' : Env) q,
      SEQ E' Si Fi q ->
      p' = p a ->
      (forall q, InE E' q -> InE (cons (p a) E) q) -> SEQ E Si Fi q) ->
  forall Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall (E' : Env) q,
   SEQ E' Si Fi q ->
   p' = extx _ A p ->
   (forall q, InE E' q -> InE (cons (extx _ A p) E) q) -> SEQ E Si Fi q.
intros A p H. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H3 _ H4). rewrite H5. rewrite H2 in H0. auto. auto.
 inversion H2.
 apply seq_cptr_l with l P; auto. intro. apply H2 with E'; auto.
 inversion H2.
 apply seq_ref_l with l P; auto. intro. apply H2 with E'; auto.
 inversion H2.
 apply seq_prop_l with p0; auto. intro. apply H2 with E'; auto.
 inversion H1.
 apply seq_and_l1 with p0 q; auto. apply seq_cons_wk with E' (extx _ A p); eauto.
 apply seq_and_l2 with p0 q; auto. apply seq_cons_wk with E' (extx _ A p); eauto.
 inversion H2.
 inversion H2.
 apply seq_or_l with p0 q; auto. apply seq_cons_wk with E' (extx _ A p); eauto.
   apply seq_cons_wk with E' (extx _ A p); eauto.
 inversion H2.
 apply seq_imp_l with p0 q; auto. apply seq_cons_wk with E' (extx _ A p); eauto.
 inversion H3.
 apply seq_allx_l with A0 P B; auto. apply seq_cons_wk with E' (extx _ A p); eauto.
 generalize H3. clear IHSEQ. cut (exists x, SEQ E Si Fi (p x)). generalize E.
   clear E A0 P x H0 H2 H3. induction H1; intros.
  destruct (H3 _ H0). rewrite H2. destruct H1. apply seq_extx_r with x; auto.
    apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (extx _ A p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (extx _ A p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p0. apply in_neq with E (extx _ A p); auto. discriminate.
    intro. apply H2; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (extx _ A p); auto. discriminate.
    apply seq_aux5 with E A p; auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (extx _ A p); auto. discriminate.
    apply seq_aux5 with E A p; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (extx _ A p); auto. discriminate.
    apply seq_aux5 with E A p; auto. apply seq_aux5 with E A p; auto.
  apply seq_imp_r. apply seq_aux5 with E A p; auto.
  apply seq_imp_l with p0 q. apply in_neq with E (extx _ A p); auto. discriminate.
    auto. apply seq_aux5 with E A p; auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A0 P B. apply in_neq with E (extx _ A p); auto.
    discriminate. apply seq_aux5 with E A p; auto.
  apply seq_extx_r with x. auto.
  destruct (H4 _ H0). rewrite H5 in H0.
    cut (forall B E0, (exists x, SEQ E0 Si Fi (p x)) -> (forall q, InE (cons (p B) E) q -> InE (cons (extx _ A p) E0) q) -> SEQ E0 Si Fi q).
    cut (forall B, SEQ (cons (p B) E) Si Fi q).
    clear A0 P H1 H2 H5. intros.
    destruct H3. apply H with x (p x) (cons (p x) E0); auto.
    apply seq_aux5 with E A p; auto.
    exists x. auto. intros. apply H2 with x; auto. inversion H5.
    generalize (inj_pairT2 _ _ _ _ _ H8). intro. inversion H6. auto.
    generalize H2. inversion H5. generalize (inj_pairT2 _ _ _ _ _ H8). intro.
    inversion H6. auto.
    apply seq_extx_l with A0 P. auto. intro. apply seq_aux5 with E A p; auto.
    intros. apply H2 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
  inversion H2. generalize (inj_pairT2 _ _ _ _ _ H6). intro. inversion H4.
    exists x. apply H0.
 apply seq_extx_l with A0 P; auto. intro. apply seq_cons_wk with E' (extx _ A p);
   eauto.
 inversion H3.
 inversion H2.
 inversion H2.
Qed.

Lemma Admissibility_of_Cut_allv :
  forall (A : Type) p,
   forall Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall (E' : Env) q,
   SEQ E' Si Fi q ->
   p' = allv _ A p ->
   (forall q, InE E' q -> InE (cons (allv _ A p) E) q) -> SEQ E Si Fi q.
intros A p. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p0; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p0 q; auto. apply seq_cons_wk with E' (allv _ A p); eauto.
 apply seq_and_l2 with p0 q; auto. apply seq_cons_wk with E' (allv _ A p); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p0 q; auto. apply seq_cons_wk with E' (allv _ A p); eauto.
   apply seq_cons_wk with E' (allv _ A p); eauto.
 inversion H1.
 apply seq_imp_l with p0 q; auto. apply seq_cons_wk with E' (allv _ A p); eauto.
 inversion H2.
 apply seq_allx_l with A0 P B; auto. apply seq_cons_wk with E' (allv _ A p); eauto.
 inversion H1.
 apply seq_extx_l with A0 P; auto. intro. apply seq_cons_wk with E' (allv _ A p);
   eauto.
 rewrite <- H2 in H3. generalize A0 p0 H1 E H H3. clear E H1 A A0 p0 p H H2 H0 H3.
   intros A p H2. induction H2; intros.
  destruct (H3 _ H). rewrite H1. apply seq_allv_r; auto. apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (allv tO A p); auto. discriminate.
    intro. apply H1; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (allv tO A p); auto. discriminate.
    intro. apply H1; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p0. apply in_neq with E (allv tO A p); auto. discriminate.
    intro. apply H1; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (allv tO A p); auto. discriminate.
    apply seq_aux6 with E A p; auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (allv tO A p); auto. discriminate.
    apply seq_aux6 with E A p; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (allv tO A p); auto. discriminate.
    apply seq_aux6 with E A p; auto. apply seq_aux6 with E A p; auto.
  apply seq_imp_r. apply seq_aux6 with E A p; auto.
  apply seq_imp_l with p0 q. apply in_neq with E (allv tO A p); auto. discriminate.
    auto. apply seq_aux6 with E A p; auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A0 P B. apply in_neq with E (allv tO A p); auto. discriminate.
    apply seq_aux6 with E A p; auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A0 P. apply in_neq with E (allv tO A p); auto.
    discriminate. intro. apply seq_aux6 with E A p; auto. intros.
    apply H1 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q. auto.
  apply seq_mu_r. auto.
 inversion H1.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_extv :
  forall (A : Type) p,
  forall Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall (E' : Env) q,
   SEQ E' Si Fi q ->
   p' = extv _ A p ->
   (forall q, InE E' q -> InE (cons (extv _ A p) E) q) -> SEQ E Si Fi q.
intros A p. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p0; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p0 q; auto. apply seq_cons_wk with E' (extv _ A p); eauto.
 apply seq_and_l2 with p0 q; auto. apply seq_cons_wk with E' (extv _ A p); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p0 q; auto. apply seq_cons_wk with E' (extv _ A p); eauto.
   apply seq_cons_wk with E' (extv _ A p); eauto.
 inversion H1.
 apply seq_imp_l with p0 q; auto. apply seq_cons_wk with E' (extv _ A p); eauto.
 inversion H2.
 apply seq_allx_l with A0 P B; auto. apply seq_cons_wk with E' (extv _ A p); eauto.
 inversion H1.
 apply seq_extx_l with A0 P; auto. intro. apply seq_cons_wk with E' (extv _ A p);
   eauto.
 inversion H2.
 rewrite <- H1 in H2. generalize H2. clear IHSEQ. cut (exists q, SEQ E Si Fi (Subst _ _ p0 q)).
   generalize A0 p0 q0 H0 E. clear E A p H H1 H2 q A0 p0 q0 H0. intros A p.
   induction 1; intros.
  destruct (H2 _ H). rewrite H1. destruct H0. apply seq_extv_r with x; auto.
    apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (extv tO A p); auto. discriminate.
    intro. apply H1; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (extv tO A p); auto. discriminate.
    intro. apply H1; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p0. apply in_neq with E (extv tO A p); auto. discriminate.
    intro. apply H1; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (extv tO A p); auto. discriminate.
    apply seq_aux7 with E A p; auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (extv tO A p); auto. discriminate.
    apply seq_aux7 with E A p; auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (extv tO A p); auto. discriminate.
    apply seq_aux7 with E A p; auto. apply seq_aux7 with E A p; auto.
  apply seq_imp_r. apply seq_aux7 with E A p; auto.
  apply seq_imp_l with p0 q. apply in_neq with E (extv tO A p); auto. discriminate.
    auto. apply seq_aux7 with E A p; auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A0 P B. apply in_neq with E (extv tO A p); auto.
    discriminate. apply seq_aux7 with E A p; auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A0 P. apply in_neq with E (extv tO A p); auto.
    discriminate. intros. apply seq_aux7 with E A p; auto. intros.
    apply H1 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q; auto.
  apply seq_mu_r. auto.
  exists q. auto.
 inversion H1.
Qed.

Lemma Admissibility_of_Cut_mu :
   forall (A : Type) (p : A -> PropX (tplus tO (tS A tO)))
     (a : A) Si Fi (E : Env) p',
   SEQ E Si Fi p' ->
   forall E' q,
   SEQ E' Si Fi q ->
   p' = mu tO A p a ->
   (forall q : PropX tO, InE E' q -> InE (cons (mu tO A p a) E) q) ->
   SEQ E Si Fi q.
intros A p a. induction 1; intros.
 apply Structural_Property_of_Sequent with E'. auto. intros.
   destruct (H2 _ H3). rewrite H4. rewrite H1 in H. auto. auto.
 inversion H1.
 apply seq_cptr_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_ref_l with l P; auto. intro. apply H1 with E'; auto.
 inversion H1.
 apply seq_prop_l with p0; auto. intro. apply H1 with E'; auto.
 inversion H2.
 apply seq_and_l1 with p0 q; auto. apply seq_cons_wk with E' (mu _ A p a); eauto.
 apply seq_and_l2 with p0 q; auto. apply seq_cons_wk with E' (mu _ A p a); eauto.
 inversion H1.
 inversion H1.
 apply seq_or_l with p0 q; auto. apply seq_cons_wk with E' (mu _ A p a); eauto.
   apply seq_cons_wk with E' (mu _ A p a); eauto.
 inversion H1.
 apply seq_imp_l with p0 q; auto. apply seq_cons_wk with E' (mu _ A p a); eauto.
 inversion H2.
 apply seq_allx_l with A0 P B; auto. apply seq_cons_wk with E' (mu _ A p a); eauto.
 inversion H1.
 apply seq_extx_l with A0 P; auto. intro. apply seq_cons_wk with E' (mu _ A p a);
   eauto.
 inversion H2.
 inversion H1.
 rewrite <- H1 in H2. generalize H2. clear IHSEQ.
   generalize A0 p0 v q H0 E H.
   clear E A p a H H1 H2 A0 p0 v q H0. intros A p v.
   induction 1; intros.
  destruct (H2 _ H). rewrite H1. apply seq_mu_r; auto.
    apply seq_env. auto.
  apply seq_cptr_r. auto.
  apply seq_cptr_l with l P. apply in_neq with E (mu tO A p v); auto. discriminate.
    intro. apply H1; auto.
  apply seq_ref_r. auto.
  apply seq_ref_l with l P. apply in_neq with E (mu tO A p v); auto. discriminate.
    intro. apply H1; auto.
  apply seq_prop_r. auto.
  apply seq_prop_l with p0. apply in_neq with E (mu tO A p v); auto. discriminate.
    intro. apply H1; auto.
  apply seq_and_r; auto.
  apply seq_and_l1 with p0 q. apply in_neq with E (mu tO A p v); auto. discriminate.
    apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
  apply seq_and_l2 with p0 q. apply in_neq with E (mu tO A p v); auto. discriminate.
    apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
  apply seq_or_r1. auto.
  apply seq_or_r2. auto.
  apply seq_or_l with p0 q. apply in_neq with E (mu tO A p v); auto. discriminate.
   apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
   apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
  apply seq_imp_r. apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
  apply seq_imp_l with p0 q. apply in_neq with E (mu tO A p v); auto. discriminate.
    auto. apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
  apply seq_allx_r. auto.
  apply seq_allx_l with A0 P B. apply in_neq with E (mu tO A p v); auto.
    discriminate. apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
  apply seq_extx_r with x. auto.
  apply seq_extx_l with A0 P. apply in_neq with E (mu tO A p v); auto.
    discriminate. intros. apply seq_aux8 with E A p v; auto; simpl; rewrite <- (ext_eq _ _ p (fun v => p v)); auto.
    intros. apply H1 with B; auto.
  apply seq_allv_r. auto.
  apply seq_extv_r with q; auto.
  apply seq_mu_r. auto.
Qed.

Lemma tl_eq_false : forall T t, tS T t = tO -> False.
intros. inversion H.
Qed.

Lemma Admissibility_of_Cut_aux : forall L (p:PropX L) Si Fi E E' q, SEQ E' Si Fi q ->
    match L as t return forall p:PropX t, Prop with
    | tO => fun p => SEQ E Si Fi p -> (forall q, InE E' q -> InE (cons p E) q) -> SEQ E Si Fi q
    | _  => fun _ => True
    end p.
induction p; intros; auto; try destruct L; auto; intros.
apply Admissibility_of_Cut_cptr with w p (cptr tO w p) E'; auto.
apply Admissibility_of_Cut_ref with w p (ref tO w p) E'; auto.
apply Admissibility_of_Cut_prop with P (prop tO P) E'; auto.
apply Admissibility_of_Cut_and with p1 p2 (p1 ./\ p2) E'; auto.
  intros. apply IHp1 with E'0; auto. rewrite <- H4. auto.
  intros. apply IHp2 with E'0; auto. rewrite <- H4. auto.
apply Admissibility_of_Cut_or with p1 p2 (p1 .\/ p2) E'; auto.
  intros. apply IHp1 with E'0; auto. rewrite <- H4. auto.
  intros. apply IHp2 with E'0; auto. rewrite <- H4. auto.
apply Admissibility_of_Cut_imp with p1 p2 (p1 .-> p2) E'; auto.
  intros. apply IHp1 with E'0; auto. rewrite <- H4. auto.
  intros. apply IHp2 with E'0; auto. rewrite <- H4. auto.
apply Admissibility_of_Cut_allx with A p (allx tO A p) E'; auto.
  intros. apply H with a E'0; auto. rewrite <- H5. auto.
apply Admissibility_of_Cut_extx with A p (extx tO A p) E'; auto.
  intros. apply H with a E'0; auto. rewrite <- H5. auto.
apply Admissibility_of_Cut_allv with A p (allv tO A p) E'; auto.
apply Admissibility_of_Cut_extv with A p (extv tO A p) E'; auto.
apply Admissibility_of_Cut_mu with A p a (mu tO A p a) E'; auto.
Qed.

Definition Admissibility_of_Cut := Admissibility_of_Cut_aux tO.

Theorem Cut_Elimination : forall E Si Fi p, SEP E Si Fi p -> SEQ E Si Fi p.
induction 1.
apply Admissibility_of_Cut with p (cons p E); auto.
apply seq_env; auto.
apply seq_cptr_r; auto.
apply seq_cptr_l with l P; auto.
apply seq_ref_r; auto.
apply seq_ref_l with l P; auto.
apply seq_prop_r; auto.
apply seq_prop_l with p; auto.
apply seq_and_r; auto.
apply seq_and_l1 with p q; auto.
apply seq_and_l2 with p q; auto.
apply seq_or_r1; auto.
apply seq_or_r2; auto.
apply seq_or_l with p q; auto.
apply seq_imp_r; auto.
apply seq_imp_l with p q; auto.
apply seq_allx_r; auto.
apply seq_allx_l with A P B; auto.
apply seq_extx_r with x; auto.
apply seq_extx_l with A P; auto.
apply seq_allv_r; auto.
apply seq_extv_r with q; auto.
apply seq_mu_r; auto.
Qed.

Theorem Normalization_of_Natural_Deduction :
  forall E Si Fi p, OK E Si Fi p -> OB E Si Fi p true.
intros.
generalize (Completeness_of_Annotated_Deductions _ _ _ _ H). clear H. intro.
generalize (Completeness_of_Sequent_Derivations_with_Cut _ _ _ _ _ H). clear H. intro.
generalize (Cut_Elimination _ _ _ _ H). clear H. intro.
generalize (Soundness_of_Sequent_Calculus _ _ _ _ H). clear H. intro.
exact H.
Qed.

Lemma Validity_of_Deductions :
  forall L Si Fi (p:PropX L), match L as t return forall p:PropX t, Prop with
   tO => fun p => OK nil Si Fi p ->
    match p in PropX t return t=tO -> Prop with
    | var _ _ _  => fun pf => False
    | lift _ _ _ => fun pf => False
    | cptr _ l P => fun pf => lookup Si l (fun S => eq_rect _ _ (P S) _ pf)
    | ref  _ l P => fun pf => lookup Fi l (fun S => eq_rect _ _ (P S) _ pf)
    | prop _ p   => fun pf => p
    | andx _ p q => fun pf => OK nil Si Fi (eq_rect _ _ p _ pf) /\ OK nil Si Fi (eq_rect _ _ q _ pf)
    | orx _ p q  => fun pf => OK nil Si Fi (eq_rect _ _ p _ pf) \/ OK nil Si Fi (eq_rect _ _ q _ pf)
    | impx _ p q => fun pf => OK nil Si Fi (eq_rect _ _ p _ pf) -> OK nil Si Fi (eq_rect _ _ q _ pf)
    | allx _ A P => fun pf => forall x, OK nil Si Fi (eq_rect _ _ (P x) _ pf)
    | extx _ A P => fun pf => exists x, OK nil Si Fi (eq_rect _ _ (P x) _ pf)
    | allv _ A p => fun pf => forall x, OK nil Si Fi (Subst _ _
      (eq_rect _ _ p _
             (trans_eq (eq_tplus _ _ _ pf)
                       (sym_eq (eq_tplus_tO _))))
      x)
    | extv _ A p => fun pf => exists x, OK nil Si Fi (Subst _ _
      (eq_rect _ _ p _
             (trans_eq (eq_tplus _ _ _ pf)
                       (sym_eq (eq_tplus_tO _))))
      x)
    | mu _ _ P v => fun pf => OK nil Si Fi (Subst _ _
                                   (eq_rect _ _ (P v)
                                               _ (trans_eq (eq_tplus _ _ _ pf)
                                                      (sym_eq (eq_tplus_tO _))))
                                                      (fun v => mu tO _ (fun v => (eq_rect _ _ (P v)
                                               _ (eq_tplus _ _ _ pf))) v))

    end (refl_equal _)
   | _ => fun p => True
   end p.
destruct p; intros; auto; try destruct L; auto; intro;
  generalize (Normalization_of_Natural_Deduction _ _ _ _ H); intro;
  generalize (Validity_of_Normal_Deductions _ _ _ _ true H0 (refl_equal _));
  intros; auto.
destruct H1. split; apply Soundness_of_Normal_Deductions with true; auto.
destruct H1. left. apply Soundness_of_Normal_Deductions with true; auto.
  right. apply Soundness_of_Normal_Deductions with true; auto.
apply ok_imp_e with p1. auto. apply ok_imp_i.
  apply Soundness_of_Normal_Deductions with true; auto.
apply Soundness_of_Normal_Deductions with true; auto.
destruct H1. exists x. apply Soundness_of_Normal_Deductions with true; auto.
apply Soundness_of_Normal_Deductions with true; auto.
destruct H1. exists x. apply Soundness_of_Normal_Deductions with true; auto.
apply Soundness_of_Normal_Deductions with true; auto.
Qed.

Definition okvalid := Validity_of_Deductions tO.

Definition Itp P Si Fi := OK nil Si Fi P.

Notation "[[ P ]]. Si , Fi" := (Itp P Si Fi) (at level 130).

Definition EXT' (t : WordTy) := forall Si Fi Fi' w, subseteq Fi Fi' ->
                                  ([[t w]]. Si,Fi) -> ([[t w]]. Si,Fi').

Definition DH Si Fi H := forall l t,
  (lookup Fi l t -> (exists w, Map.lookup H l w /\ [[t w]]. Si,Fi) /\ EXT' t)
    /\ (Map.in_dom H l -> in_dom Fi l).

Definition ItP := (fun (a:Assertion) Si S => match S with (H, R) =>
                     exists Fi, exists H1, exists H2,
                        Map.disjoint H1 H2 /\ Map.merge H1 H2 = H
                     /\ ([[a (H1,R)]]. Si,Fi) /\ DH Si Fi H2
                   end).

Notation "[[ a ]] Si" := (fun S => match S with (H, R) =>
                     exists Fi, exists H1, exists H2,
                        Map.disjoint H1 H2 /\ Map.merge H1 H2 = H
                     /\ ([[a (H1,R)]]. Si,Fi) /\ DH Si Fi H2
                   end) (at level 130).

Notation "p ==>xcap q" :=
  (forall Si S, ItP p Si S -> ItP q Si S)
  (at level 130, right associativity).

Notation "p ==> q" := (p ==>xcap q) (at level 130, right associativity).

Inductive WFiseq : CdHpSpec -> Assertion -> InstrSeq -> Prop :=
  | wfiseq : forall Si a c I a',
               (forall Si s, ItP a Si s -> exists s',
                      Next c s s' /\ ItP a' Si s') ->
               WFiseq Si a' I                          -> WFiseq Si a (iseq c I)
  | wfbgt  : forall Si a rs rt l I a' a'',
               ((fun s => <<_R s rs <= _R s rt >> ./\ a s) ==> a'') ->
               ((fun s => <<_R s rs >  _R s rt >> ./\ a s) ==> a') ->
               lookup Si l a' -> WFiseq Si a'' I  -> WFiseq Si a (bgt rs rt l I)
  | wfbgti : forall Si a rs w l I a' a'',
               ((fun s => << _R s rs <= w >> ./\ a s) ==> a'') ->
               ((fun s => << _R s rs >  w >> ./\ a s) ==> a') ->
               lookup Si l a' -> WFiseq Si a'' I  -> WFiseq Si a (bgti rs w l I)
  | wfjd   : forall Si a l a', lookup Si l a' -> (a ==> a')
                                                           -> WFiseq Si a (jd l)
  | wfjmp  : forall Si a r,
               (a ==> fun S => extv _ _ (eq_rect _ _
                                 (cptr _ (_R S r)
                                       (fun S => (var _ _ S)) ./\ var _ _ S)
                                 _ (eq_tplus_tO _)))
                                                          -> WFiseq Si a (jmp r)
  | wfecp  : forall Si a l a' a'' I,
               WFiseq Si a' I -> lookup Si l a'' ->
               ((fun s => cptr _ l a'' ./\ a s) ==> a')         -> WFiseq Si a I.

Notation "Si |-xcap{ a } i" := (WFiseq Si a i) (at level 130).
Notation "Si |-{ a } i" := (Si |-xcap{ a}i) (at level 130).

Inductive WFcode : CdHpSpec -> CodeHeap -> CdHpSpec -> Prop :=
  | wfcdhp : forall Si Si' C,
               (forall l a, lookup Si' l a
                  -> exists I : _, Map.lookup C l I /\ (Si' |-{ a}I))
                                                              -> WFcode Si C Si'
  | wflink :
      forall Si1 Si2 Si'1 Si'2 C1 C2 Si Si',
      WFcode Si1 C1 Si'1 ->
      WFcode Si2 C2 Si'2 ->
      Map.disjoint C1 C2 ->
      subseteq Si1 (merge Si1 Si2) ->
      subseteq Si2 (merge Si1 Si2) ->
      (forall l, merge Si1 Si2 l = Si l) ->
      (forall l, merge Si'1 Si'2 l = Si' l)   -> WFcode Si (Map.merge C1 C2) Si'.

Notation "Si |-xcap C :: Si'" := (WFcode Si C Si') (at level 130).
Notation "Si |- C :: Si'" := (Si |-xcap C :: Si') (at level 130).

Definition WFprogram Si a P :=
  (Si |- (fun P : Program => let (C, p) := P in let (_, _) := p in C) P :: Si) /\
  (Si |-{ a}(fun P : Program => let (_, p) := P in let (_, i) := p in i) P) /\
  ItP a Si
    ((fun P : Program => let (_, p) := P in let (s, _) := p in s) P).

Notation "Si |-xcap[ a ] P" := (WFprogram Si a P) (at level 130).
Notation "Si |-[ a ] P" := (Si |-xcap[ a]P) (at level 130).

Lemma InstrSeqWeakening : forall Si Si' a a' I,
                            (Si |-{a'} I) -> (subseteq Si Si') -> (a ==> a')
                                                                -> (Si' |-{a} I).
intros; move H after Si'; generalize a H0 H1; clear a H0 H1.
induction H.
intros.
apply wfiseq with a'.
auto. apply IHWFiseq. auto. auto.
intros.
apply wfbgt with a' a''.
intros.
apply H.
destruct S. destruct H5. destruct H5. destruct H5. destruct H5. destruct H6.
destruct H7.
destruct (H4 Si0 (Map.merge x0 x1,r)).
exists x. exists x0. exists x1. split; auto. split; auto. split; auto.
intros. destruct (okvalid _ _ _ H7). clear H7. simpl in *. auto.
destruct H9. destruct H9. destruct H9. destruct H10. destruct H11.
exists x2. exists x3. exists x4. split; auto. split. rewrite <- H6. auto.
split; auto. intros. unfold Itp.
destruct (okvalid _ _ _ H7); simpl in *.
generalize (okvalid _ _ _ H13); intro.
apply ok_and_i; auto.
apply ok_prop_i; auto.
intros. apply H0.
destruct S. destruct H5. destruct H5. destruct H5. destruct H5. destruct H6.
destruct H7.
destruct (H4 Si0 (Map.merge x0 x1, r)).
exists x. exists x0. exists x1. split; auto; split; auto. split; auto.
intros. destruct (okvalid _ _ _ H7). clear H7. simpl in *. auto.
destruct H9. destruct H9. destruct H9. destruct H10. destruct H11.
exists x2. exists x3. exists x4. split; auto. split. rewrite <- H6. auto.
split; auto. intros. unfold Itp.
destruct (okvalid _ _ _ H7); simpl in *.
generalize (okvalid _ _ _ H13); intro.
apply ok_and_i; auto.
apply ok_prop_i; auto.
unfold subseteq in H3. unfold lookup in *. generalize (H3 l).
destruct (Si l); auto. rewrite H1. auto. inversion H1.
auto.
intros.
apply wfbgti with a' a''.
intros.
apply H.
destruct S. destruct H5. destruct H5. destruct H5. destruct H5. destruct H6.
destruct H7.
destruct (H4 Si0 (Map.merge x0 x1,r)).
exists x. exists x0. exists x1. split; auto. split; auto. split; auto.
intros. destruct (okvalid _ _ _ H7). clear H7. simpl in *. auto.
destruct H9. destruct H9. destruct H9. destruct H10. destruct H11.
exists x2. exists x3. exists x4. split; auto. split. rewrite <- H6. auto.
split; auto. intros. unfold Itp.
destruct (okvalid _ _ _ H7); simpl in *.
generalize (okvalid _ _ _ H13); intro.
apply ok_and_i; auto.
apply ok_prop_i; auto.
intros. apply H0.
destruct S. destruct H5. destruct H5. destruct H5. destruct H5. destruct H6.
destruct H7.
destruct (H4 Si0 (Map.merge x0 x1, r)).
exists x. exists x0. exists x1. split; auto; split; auto. split; auto.
intros. destruct (okvalid _ _ _ H7). clear H7. simpl in *. auto.
destruct H9. destruct H9. destruct H9. destruct H10. destruct H11.
exists x2. exists x3. exists x4. split; auto. split. rewrite <- H6. auto.
split; auto. intros. unfold Itp.
destruct (okvalid _ _ _ H7); simpl in *.
generalize (okvalid _ _ _ H13); intro.
apply ok_and_i; auto.
apply ok_prop_i; auto.
unfold subseteq in H3. unfold lookup in *. generalize (H3 l).
destruct (Si l); auto. rewrite H1. auto. inversion H1.
auto.
intros.
apply wfjd with a'.
unfold lookup.
unfold lookup in H.
generalize (H1 l).
rewrite H.
auto.
auto.
intros.
apply wfjmp.
intros.
simpl.
unfold Itp.
generalize (H Si0 S (H1 Si0 S H2)).
simpl.
auto.
intros.
apply wfecp with l a' a''.
apply IHWFiseq.
auto.
auto.
inversion H0.
generalize (H2 l).
rewrite H5.
auto.
intros.
unfold lookup.
auto.
intros.
apply H1.
auto.
destruct S. destruct (H3 Si0 (o,r)).
destruct H4. destruct H4. destruct H4. destruct H4. destruct H5.
destruct H6.
exists x. exists x0. exists x1. split; auto. split; auto. split; auto.
intros. destruct (okvalid _ _ _ H6). clear H6. simpl in *. auto.
destruct H5. destruct H5. destruct H5. destruct H6. destruct H7.
exists x. exists x0. exists x1. split; auto. split; auto. split; auto.
intros.
unfold Itp. apply ok_and_i; auto.
destruct H4. destruct H4. destruct H4. destruct H4. destruct H9.
destruct H10.
destruct (okvalid _ _ _ H10). simpl in *.
apply ok_cptr_i.
generalize (okvalid _ _ _ H12). simpl.
rewrite (ext_eq _ _ (fun S => a'' S) a''); auto.
Qed.

Lemma CodeHeapTyping :
 forall C Si l a,
 (Si |- C :: Si) ->
 lookup Si l a -> exists I : _, Map.lookup C l I /\ (Si |-{ a}I).
intros.
generalize l a H0.
clear l a H0.
induction H.
intros.
destruct (H l a H0).
split with x.
auto.
intros.
unfold lookup in H6.
rewrite <- (H5 l) in H6.
destruct (mergelookup H6).
destruct (IHWFcode1 l a H7).
split with x.
destruct H8.
split.
unfold Map.lookup, Map.merge in |- *.
unfold Map.lookup in H8.
rewrite H8.
auto.
apply InstrSeqWeakening with Si'1 a.
auto.
unfold subseteq in |- *.
intro.
rewrite <- (H5 a0).
unfold merge in |- *.
destruct (Si'1 a0).
auto.
auto.
auto.
destruct (IHWFcode2 l a H7).
destruct H8.
split with x.
split.
unfold Map.lookup, Map.merge in |- *.
unfold Map.disjoint in H1.
generalize (H1 l).
destruct (C1 l).
unfold Map.lookup in H8.
rewrite H8.
tauto.
unfold Map.lookup in H8.
rewrite H8.
auto.
apply InstrSeqWeakening with Si'2 a.
auto.
unfold subseteq in |- *.
intro.
rewrite <- (H5 a0).
unfold merge in |- *.
generalize (IHWFcode2 a0).
unfold lookup.
destruct (Si'2 a0).
intros.
generalize (IHWFcode1 a0).
unfold lookup in |- *.
destruct (Si'1 a0).
intros.
destruct (H11 a2).
auto.
destruct H12.
destruct (H10 a1).
auto.
destruct H14.
unfold Map.lookup in H12.
unfold Map.lookup in H14.
generalize (H1 a0).
rewrite H12.
rewrite H14.
tauto.
auto.
auto.
auto.
Qed.

Theorem Soundness :
 forall Si a P,
 (Si |-[ a]P) ->
 exists a' : _, (exists P' : _, (P |--> P') /\ (Si |-[ a']P')).
intros.
destruct P.
destruct p.
destruct H.
destruct H0.
induction H0.
exists a'.
destruct (H0 Si s); auto. destruct H3.
exists (c, (x, I)). split.
apply stp_iseq; auto.
split; auto.
destruct s.
destruct (le_or_gt (r rt) (r rs)).
split with a''.
split with (c, (h, r, I)).
split.
apply stp_bgt'; auto.
split; auto.
split; auto.
apply H0.
destruct H1. destruct H1. destruct H1. destruct H1. destruct H6. destruct H7.
exists x. exists x0. exists x1. split; auto. split; auto. split; auto.
intros. unfold Itp. apply ok_and_i; auto. apply ok_prop_i; auto.
destruct (CodeHeapTyping c Si l a'); auto.
split with a'. split with (c, (h, r, x)). split.
apply stp_bgt; tauto.
split; auto.
split; try tauto.
apply H2.
destruct H1. destruct H1. destruct H1. destruct H1. destruct H7. destruct H8.
exists x0. exists x1. exists x2. split; auto. split; auto. split; auto.
intros. unfold Itp. apply ok_and_i; auto. apply ok_prop_i; auto.
destruct s.
destruct (le_or_gt w (r rs)).
split with a''.
split with (c, (h, r, I)).
split.
apply stp_bgti'; auto.
split; auto.
split; auto.
apply H0.
destruct H1. destruct H1. destruct H1. destruct H1. destruct H6. destruct H7.
exists x. exists x0. exists x1. split; auto. split; auto. split; auto.
intros; unfold Itp. apply ok_and_i; auto. apply ok_prop_i; auto.
destruct (CodeHeapTyping c Si l a'); auto.
split with a'. split with (c, (h, r, x)). split.
apply stp_bgti; tauto.
split; auto.
split; try tauto.
apply H2.
destruct H1. destruct H1. destruct H1. destruct H1. destruct H7. destruct H8.
exists x0. exists x1. exists x2. split; auto. split; auto. split; auto.
intros; unfold Itp. apply ok_and_i; auto. apply ok_prop_i; auto.
split with a'.
destruct (CodeHeapTyping c Si l a'); auto.
destruct H3.
split with (c, (s, x)).
split.
apply stp_jd; auto.
split; auto.
destruct s.
destruct (H0 Si (h,r0) H1). clear H0. destruct H2. destruct H0. destruct H0.
destruct H2. destruct H3. simpl in H3.
destruct (okvalid _ _ _ H3). clear H3. simpl in *.
destruct (okvalid _ _ _ H5). clear H5. simpl in *.
generalize (okvalid _ _ _ H3); clear H3; intro. simpl in *.
rewrite (ext_eq _ _ (fun S => x2 S) x2) in H3; auto.
destruct (CodeHeapTyping c Si (r0 r) x2); auto.
destruct H5.
exists x2.
split with (c, (h, r0, x3)).
split.
apply stp_jmp; auto.
split; auto.
split; auto.
exists x. exists x0. exists x1. auto.
intros.
apply IHWFiseq; auto.
apply H3. destruct s. destruct H1. destruct H1. destruct H1. destruct H1.
destruct H4. destruct H5.
exists x. exists x0. exists x1. split; auto. split; auto. split; auto.
unfold Itp. apply ok_and_i; auto. apply ok_cptr_i; auto.
Qed.

